################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Definition of CircuitRunner protocol and ABC for implementing it."""
from abc import ABC, abstractmethod
from typing import List, Optional, Protocol, Sequence, Union

from ..circuits import Circuit
from ..distributions import MeasurementOutcomeDistribution
from ..measurements import Measurements


class CircuitRunner(Protocol):
    """Protocol for objects able to run circuits and collect measurements.

    This protocol is a pure interface and does not contain any default
    implementations. See BaseCircuitRunner ABC that can be extended to
    implement simple CircuitRunner.

    Note:
        This protocol was previously known as QuantumBackend.
    """

    def run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """Run circuit given number of times and return obtained measurements.

        Args:
            circuit: a circuit to be run.
            n_samples: minimum number of times the circuit should be run. This
              is sometimes called "number of shots" in other SDKs.
        Returns:
            A Measurements object containing *at least* n_samples.
        Raises:
            ValueError: if n_samples is not positive.

        Note:
            For some runners, the returned number of measurements might be
            actually larger than n_samples. Runner - independent code should
            take it into account and not rely on the assumption
            len(measurement.bitstrings) == n_samples
        """

    def run_batch_and_measure(
        self, circuits_batch: Sequence[Circuit], n_samples: Union[int, Sequence[int]]
    ) -> List[Measurements]:
        """Run multiple circuits and return measurements for each circuit.

        Essentially, this method performs the same as `run_and_measure` for
        a single circuit. However, some runners may be able to submit multiple
        circuits at once (thus possibly reducing costs), or run multiple
        circuits in parallel (thus increasing performance). For this reasons,
        usage of run_batch_and_measure is always recommended over
        run_and_measure if one needs to run multiple circuits.

        Args:
            circuits_batch: sequence of circuits to be run.
            n_samples: minimum number of times each circuit should be run. If
              a single integer N is passed, each circuit will be run at least
              N times. If a sequence of integers is passed, it must be of the
              same length as circuits_batch and i-th number is treated as
              minimum number of runs for i-th circuit.
        Returns:
            A list of measurements such that i-th object corresponds to i-th
            circuit.
        Raises:
            ValueError:
            - for integral n_samples if it is not positive
            - if any element of n_samples sequence is not positive, or n_samples
              has length differing from circuits_batch.

        Note:
            Similarly as with `run_and_measure` method, clients of circuit
            runners can only assume number of returned samples is greater
            or equal to the requested one. Furthermore, clients should not
            assume that each cicruict has been run the same number of times.
        """

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int]
    ) -> MeasurementOutcomeDistribution:
        """Get a distribution of measurement outcomes from a given circuit.

        This is essentially the same as `run_and_measure`, except that
        a frequency of obtained outcomes is returned rather than a
        Measurements object.

        If `n_samples` is equal to None, supported samplers may returned
        exact distribution instead of an empirical one.
        """

    @property
    def n_jobs_executed(self) -> int:
        """Number of jobs executed by this runner.

        For circuit runners wrapping external resources or simulators,
        this number determines how many calls were made to the external
        library or API.
        """

    @property
    def n_circuits_executed(self) -> int:
        """Total number of circuits executed by this runner.

        Note:
            This number is counted with repetitions. For instance, if the
            same circuit has to be run multiple times to obtain required number
            of samples.
        """


class BaseCircuitRunner(ABC, CircuitRunner):
    """ABC for implementing simple circuit runners.

    To implement this ABC, override at least `_run_and_measure` method.
    In addition, if your runner supports more sophisticated way of
    running multiple circuits in a batch, you can override the
    `_run_batch_and_measure` method.

    Note:
        For purpose of counting executed jobs and circuits, this ABC
        makes assumption that each call to _run_and_measure constitutes
        a single job. If this assumption is not valid, you should avoid
        inheriting this class and instead implement CircuitRunner protocol
        separately.
    """

    def __init__(self):
        super().__init__()
        self._n_circuits_executed = 0
        self._n_jobs_executed = 0

    def run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """Run circuit given number of times and return obtained measurements.

        See docstrings of CircuitRunner protocol for exact description of
        parameters.

        Note:
            Subclasses of `BaseCircuitRunner` ABC should not override this method.
            Instead, implement `_run_and_measure`.
        """
        if n_samples <= 0:
            raise ValueError(f"Number of samples has to be positive, got {n_samples}")
        result = self._run_and_measure(circuit, n_samples)
        self._n_circuits_executed += 1
        self._n_jobs_executed += 1
        return result

    def run_batch_and_measure(
        self, circuits_batch: Sequence[Circuit], n_samples: Union[int, Sequence[int]]
    ) -> List[Measurements]:
        """Run multiple circuits and return measurements for each circuit.

        See docstrings of CircuitRunner protocol for exact description of
        parameters.

        Note:
            Subclasses of `BaseCircuitRunner` ABC should not override this method.
            Instead, implement `_run_and_measure` method and, if you have a
            dedicated way of running circuits' batch, `_run_batch_and_measure`
            method.
        """
        samples_per_circuit = (
            len(circuits_batch) * [n_samples]
            if isinstance(n_samples, int)
            else n_samples
        )
        if len(samples_per_circuit) != len(circuits_batch):
            raise ValueError(
                "Number of samples has to be an integer or a sequence of length "
                "equal to the length of batch. Length of batch: "
                f"{len(circuits_batch)}, length of n_samples: "
                f"{len(samples_per_circuit)}."
            )
        if any(n <= 0 for n in samples_per_circuit):
            raise ValueError(
                f"All numbers of samples have to be positive. Got: {n_samples}"
            )

        return self._run_batch_and_measure(circuits_batch, samples_per_circuit)

    @abstractmethod
    def _run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """Underlying implementation of running and measuring single circuit.

        Implementations of this method can assume that n_samples is positive.
        Furthermore, implementations of this method should not modify
        `self._n_jobs_executed` and `self._n_circuits_executed` fields.
        """

    def _run_batch_and_measure(
        self, batch: Sequence[Circuit], samples_per_circuit: Sequence[int]
    ):
        """Underlying implementation of running and measuring multiple circuits.

        Implementations of this method can assume that samples_per_circuit has
        the same length as `batch`, and each element in `samples_per_circuit`
        is positive.

        The basic implementation just runs `run_and_measure` for each circuit.
        It should only be overriden if runner implements some dedicated logic
        for running circuits' batch.
        """
        return [
            self.run_and_measure(circuit, n)
            for circuit, n in zip(batch, samples_per_circuit)
        ]

    @property
    def n_jobs_executed(self) -> int:
        """Number of executed jobs."""
        return self._n_jobs_executed

    @property
    def n_circuits_executed(self) -> int:
        """Number of executed circuits."""
        return self._n_circuits_executed

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int]
    ) -> MeasurementOutcomeDistribution:
        """Get a distribution of measurement outcomes from a given circuit.

        Args:
            circuit: circuit to be sampled
            n_samples: number of times `circuit` should be sampled
        Returns:
            An object representing empirical distribution of measured bitstrings
        Raises:
            ValueError: if n_samples is None or n_samples is not positive.
        """
        if n_samples is None:
            raise ValueError(
                "This runner needs n_samples to compute measurement outcome"
                "distribution"
            )
        measurements = self.run_and_measure(circuit, n_samples)
        return measurements.get_distribution()



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
import warnings
from collections import OrderedDict
from itertools import chain, product
from numbers import Number
from typing import (
    Any,
    Dict,
    FrozenSet,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import numpy as np

from ..circuits import Circuit, Operation, builtin_gate_by_name

PauliRepresentation = Union["PauliTerm", "PauliSum"]

ALLOWED_OPERATORS = ["X", "Y", "Z", "I"]

# Use ASCII values instead of actual strings
# as keys to allow for commutativity
OPERATOR_MAP = {
    ord("X") + ord("Y"): "Z",
    ord("Y") + ord("Z"): "X",
    ord("X") + ord("Z"): "Y",
}

COEFF_MAP = {
    "XY": 1.0j,
    "XZ": -1.0j,
    "YX": -1.0j,
    "YZ": 1.0j,
    "ZX": 1.0j,
    "ZY": -1.0j,
}

HASH_PRECISION = 1e6


def _efficient_exponentiation(
    pauli_rep: PauliRepresentation, power: int
) -> PauliRepresentation:
    """Efficiently exponentiate PauliTerm or PauliSum.

    This function assumes validation checks have already been done from parent routine.
    """
    if power == 0:
        return type(pauli_rep).identity()

    if power % 2 == 1:
        return pauli_rep * _efficient_exponentiation(pauli_rep, power - 1)

    intermediate_result = _efficient_exponentiation(pauli_rep, power // 2)
    return intermediate_result * intermediate_result


def _validate_type(object: Any) -> None:
    if not isinstance(object, (PauliSum, PauliTerm, int, float, complex)):
        raise TypeError(
            f"Can't carry out operation with object of type {type(object)}."
        )


def _is_in_brackets(string: str) -> bool:
    return string.startswith("(") and string.endswith(")")


def _parse_complex(complex_str: str) -> complex:
    value = complex(complex_str.replace(" ", ""))
    if value.real != 0 and value.imag != 0 and not _is_in_brackets(complex_str):
        raise ValueError(
            "Complex number with nonzero real and imaginary part has to be "
            "enclosed in bracket."
        )
    return value


def _parse_operator(op_str: str) -> Tuple[int, str]:
    match = re.match(r"([XYZI])([0-9]+)$", op_str, re.I)

    if not match:
        raise ValueError("Badly formatted string representation passed.")

    return int(match.group(2)), match.group(1).upper()


def _parse_operators_and_coefficient(
    term_str: str,
) -> Tuple[Optional[complex], Dict[int, str]]:
    parts = re.split(r"\ *\*\ *", term_str.strip(" "))
    try:
        coef = _parse_complex(parts[0])
        operators_strs = parts[1:]
    except ValueError:
        coef = None
        operators_strs = parts

    operators_dict = dict([_parse_operator(op_str) for op_str in operators_strs])

    if len(operators_dict) != len(operators_strs):
        raise ValueError("Duplicate qubit index in a term detected.")

    return coef, dict([_parse_operator(op_str) for op_str in operators_strs])


class PauliTerm:
    """Representation of a single Pauli Term.

    If coefficient is not provided neither directly nor in the string repr,
    it defaults to 1.0
    """

    def __init__(
        self,
        operator: Union[str, Dict[int, str]],
        coefficient: Optional[complex] = None,
    ):
        if isinstance(operator, str):
            _parsed_coefficient, operator = _parse_operators_and_coefficient(operator)
            if _parsed_coefficient is not None and coefficient is not None:
                raise ValueError(
                    "Coefficient can be provided either in an argument or string "
                    "representation (but not both)."
                )
            if _parsed_coefficient is not None:
                coefficient = _parsed_coefficient

        if not all([qubit_idx >= 0 for qubit_idx in operator]):
            raise ValueError(
                "Invalid qubit index in dictionary keys. "
                "Make sure all qubit indices are non-negative integers."
            )

        # Verify value values
        if not all([op in ALLOWED_OPERATORS for op in operator.values()]):
            raise ValueError(
                "Invalid operators in dictionary. "
                f"Allowed ones are {ALLOWED_OPERATORS}."
            )

        self._ops: Dict[int, str] = {
            idx: op for idx, op in operator.items() if op != "I"
        }
        self.coefficient = 1.0 if coefficient is None else coefficient

    @staticmethod
    def from_iterable(
        terms: Iterable[Tuple[str, int]], coefficient: complex = 1.0
    ) -> "PauliTerm":
        """Construct PauliTerm from a list of operators.

        A slightly more efficient constructor when all the elements of the term are
        known beforehand. Users should employ this function instead of creating
        individual terms and multiplying.
        """

        ############### Some checks on input first ###############
        if terms:
            _, idx_list = zip(*terms)

            if not all([isinstance(op, tuple) for op in terms]):
                raise ValueError(
                    "The list can only contain tuples of the form (index, op). If you "
                    "want to initialize from strings, use PauliTerm's constructor."
                )

            if len(set(idx_list)) != len(idx_list):
                raise ValueError(
                    "Duplicate indices used in list. Manually create terms"
                    "and multiply them instead."
                )

        ##########################################################

        result_dict = {idx: op for op, idx in terms if op != "I"}

        return PauliTerm(result_dict, coefficient)

    @staticmethod
    def identity() -> "PauliTerm":
        return PauliTerm("I0", 1.0)

    def copy(self, new_coefficient: Optional[complex] = None) -> "PauliTerm":
        """Copy PauliTerm, possibly changing its coefficient to a new one.

        The created copy is deep, in particular internal dictionary storing map
        from qubit indices to operators is also copied.
        """
        if new_coefficient is None:
            new_coefficient = self.coefficient

        return PauliTerm(operator=self._ops, coefficient=new_coefficient)

    @property
    def qubits(self) -> Set[int]:
        """The list of qubit indices associated with this term."""
        return set(self._ops.keys())

    @property
    def is_ising(self) -> bool:
        """True iff this term is Ising Operator (i.e. contains no X or Y operators)"""

        return set(self._ops.values()) == {"Z"} or self.is_constant

    @property
    def circuit(self) -> Circuit:
        """Circuit implementing this Pauli term.

        For efficiency constructed circuit is cached after the first invocation of
        this property.
        """
        if not hasattr(self, "_circuit"):
            self._circuit = Circuit(
                [cast(Operation, builtin_gate_by_name(op)(index)) for op, index in self]
            )

        return self._circuit

    @property
    def operations(self) -> FrozenSet[Tuple[int, str]]:
        return frozenset(self._ops.items())

    def __len__(self) -> int:
        return len(self._ops)

    def __getitem__(self, i: int) -> str:
        if i not in self._ops:
            warnings.warn("Index not in term! Identity is returned.", UserWarning)

        return self._ops.get(i, "I")

    def __iter__(self) -> Iterator[Tuple[str, int]]:
        for i in self.qubits:
            yield self[i], i

    def __hash__(self) -> int:
        if isinstance(self.coefficient, complex):
            coefficient_real = self.coefficient.real
            coefficient_imag = self.coefficient.imag
        else:
            coefficient_real = self.coefficient
            coefficient_imag = 0
        return hash(
            (
                round(coefficient_real * HASH_PRECISION),
                round(coefficient_imag * HASH_PRECISION),
                self.operations,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (PauliSum, PauliTerm, int, float, complex)):
            raise TypeError(f"Can't compare with object of type {type(other)}")

        if isinstance(other, (int, float, complex)):
            return self == PauliTerm("I0", other)

        if isinstance(other, PauliSum):
            return other == self

        cast_other = cast(PauliTerm, other)
        return np.allclose(self.coefficient, cast_other.coefficient) and (
            np.allclose(self.coefficient, 0) or self.operations == cast_other.operations
        )

    def __add__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        _validate_type(other)

        if isinstance(other, PauliSum):
            return other + self

        if isinstance(other, PauliTerm):
            return PauliSum([self, other]).simplify()

        return self + PauliTerm("I0", other)

    def __radd__(self, other: complex) -> "PauliSum":
        return self + PauliTerm("I0", other)

    def __sub__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        return self + -1.0 * other

    def __rsub__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        return other + -1.0 * self

    def _multiply_by_operator(self, op: str, index: int) -> "PauliTerm":
        result_coeff = self.coefficient
        result_ops = self._ops.copy()

        if index not in result_ops:
            # Case 1: qubit not used yet
            result_ops[index] = op
        elif result_ops[index] == op:
            # Case 2: equal operators cancel
            del result_ops[index]
        elif op in ["X", "Y", "Z"]:
            # Case 3: different pauli ops return third op
            result_ops[index] = OPERATOR_MAP[ord(self[index]) + ord(op)]
            result_coeff *= COEFF_MAP[self[index] + op]
        else:
            raise ValueError(f"Unsupported operation. Got {op}.")

        return PauliTerm(result_ops, result_coeff)

    def __mul__(
        self, other: Union[PauliRepresentation, complex]
    ) -> PauliRepresentation:
        """Multiply this Pauli Term with another PauliTerm, PauliSum, or number.

        This method performs simplifications according to Pauli Algebra rules.
        """
        _validate_type(other)

        if isinstance(other, PauliSum):
            return (PauliSum([self]) * other).simplify()
        elif isinstance(other, PauliTerm):
            result_term = self.copy(new_coefficient=1)

            for op, index in other:
                if op != "I":
                    result_term = result_term._multiply_by_operator(op, index)

            new_coeff = self.coefficient * other.coefficient
            return result_term.copy(new_coefficient=result_term.coefficient * new_coeff)

        return self.copy(self.coefficient * complex(other))

    def __rmul__(self, other: complex) -> "PauliTerm":
        result = self * other
        assert isinstance(result, PauliTerm)
        return result

    def __truediv__(self, other: complex) -> "PauliTerm":
        result = self * (1.0 / other)
        assert isinstance(result, PauliTerm)
        return result

    def __pow__(self, power: int) -> "PauliTerm":
        """Raise this PauliTerm to integral power."""
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")

        return cast(PauliTerm, _efficient_exponentiation(self.copy(), power))

    def __repr__(self) -> str:
        term_strs = [f"{self[index]}{index}" for index in self._ops]

        if len(term_strs) == 0:
            term_strs.append("I")

        return f"{self.coefficient}*{'*'.join(term_strs)}"

    @property
    def terms(self) -> List["PauliTerm"]:
        return [self]

    @property
    def is_constant(self) -> bool:
        return self._ops == {}

    @property
    def n_qubits(self) -> int:
        """Number of qubits used in this PauliTerm.

        Follows the convention of openfermion's `count_qubits`.
        Note that this is different from the number of operations. For example,
        PauliTerm("Z0*Z3").n_qubits = 4, but len(PauliTerm("Z0*Z3").qubits) = 2.

        """
        return 0 if self.is_constant else max(self.qubits) + 1


class PauliSum:
    def __init__(self, terms: Optional[Union[str, Sequence[PauliTerm]]] = None):
        if isinstance(terms, str):
            terms = [PauliTerm(s.strip()) for s in re.split(r"\+(?![^(]*\))", terms)]
        if terms is None:
            # If no terms is given, the PauliSum has a value of zero.
            terms = []

        if not (
            isinstance(terms, Sequence)
            and all([isinstance(term, PauliTerm) for term in terms])
        ):
            raise ValueError(
                "PauliSums can be constructed only from Sequences of PauliTerms."
            )
        self.terms: Sequence[PauliTerm] = cast(Sequence[PauliTerm], terms)

    def __len__(self) -> int:
        return len(self.terms)

    def __getitem__(self, idx: int) -> PauliTerm:
        return self.terms[idx]

    def __iter__(self) -> Iterator[PauliTerm]:
        return self.terms.__iter__()

    @property
    def qubits(self) -> Set[int]:
        return set(chain.from_iterable([term.qubits for term in self.terms]))

    @property
    def is_ising(self) -> bool:
        """Returns whether the full operator represents an Ising model."""
        if not hasattr(self, "_is_ising"):
            self._is_ising = all([term.is_ising for term in self.terms])
        return self._is_ising

    @property
    def circuits(self) -> List[Circuit]:
        if not hasattr(self, "_circuits"):
            self._circuits = [term.circuit for term in self.terms]

        return self._circuits

    @staticmethod
    def identity() -> "PauliSum":
        return PauliSum([PauliTerm.identity()])

    def __eq__(self, other: object) -> bool:
        _validate_type(other)

        if isinstance(other, (int, float, complex)):
            constant_term = PauliTerm("I0", complex(other))
            return self == PauliSum([constant_term])

        if isinstance(other, PauliTerm):
            if len(self) == 0:
                return np.allclose(other.coefficient, 0)
            return self == PauliSum([other])

        other = cast(PauliSum, other)
        if len(self) != len(other):
            return False

        return set(self.terms) == set(other.terms)

    def __add__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        _validate_type(other)

        if isinstance(other, PauliTerm):
            other = PauliSum([other])
        elif isinstance(other, (int, float, complex)):
            other = PauliSum([PauliTerm("I0", other)])

        new_op = PauliSum([term.copy() for term in chain(self.terms, other.terms)])

        return new_op.simplify()

    def __radd__(self, other: complex) -> "PauliSum":
        assert isinstance(other, (int, float, complex))
        return self + other

    def __sub__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        # No need to check for types here since it will be
        # carried out in the __add__ function
        return self + -1.0 * other

    def __rsub__(self, other: complex) -> "PauliSum":
        return other + -1.0 * self

    def __mul__(self, other: Union[PauliRepresentation, complex]) -> "PauliSum":
        _validate_type(other)

        other_terms = (
            other.terms
            if isinstance(other, PauliSum)
            else [cast(PauliTerm, PauliTerm.identity() * other)]
        )

        new_paulisum = PauliSum(
            cast(
                Sequence[PauliTerm],
                [
                    left_term * right_term
                    for left_term, right_term in product(self.terms, other_terms)
                ],
            )
        )

        return new_paulisum.simplify()

    def __rmul__(self, other: complex) -> "PauliSum":
        assert isinstance(other, (int, float, complex))

        new_terms = [cast(PauliTerm, term.copy() * other) for term in self.terms]

        return PauliSum(new_terms).simplify()

    def __truediv__(self, other: complex) -> "PauliSum":
        return self * (1.0 / other)

    def __pow__(self, power: int) -> "PauliSum":
        if not isinstance(power, int) or power < 0:
            raise ValueError(f"Power must be a non-negative integer. Got {power}.")

        return cast(PauliSum, _efficient_exponentiation(self, power))

    def simplify(self) -> "PauliSum":
        like_terms: Dict[Hashable, List[PauliTerm]] = OrderedDict()
        for term in self.terms:
            key = term.operations
            if key in like_terms:
                like_terms[key].append(term)
            else:
                like_terms[key] = [term]

        terms = []
        for term_list in like_terms.values():
            first_term = term_list[0]
            if len(term_list) == 1 and not np.isclose(first_term.coefficient, 0.0):
                terms.append(first_term)
            else:
                coeff = sum(t.coefficient for t in term_list)
                if not np.isclose(coeff, 0.0):  # type: ignore
                    terms.append(term_list[0].copy(new_coefficient=coeff))
        return PauliSum(terms)

    def __repr__(self):
        if len(self) == 0:
            zero_identity_term = PauliTerm("I0", 0)
            return str(zero_identity_term)
        return " + ".join([str(term) for term in self.terms])

    def __hash__(self):
        return hash(tuple(self.terms))

    @property
    def is_constant(self) -> bool:
        return len(self.terms) == 0 or all([term.is_constant for term in self.terms])

    @property
    def constant_term(self) -> Union[Number, complex]:
        return sum([term.coefficient for term in self.terms if term.is_constant])

    @property
    def n_qubits(self) -> int:
        """Number of qubits used in this PauliSum.

        Follows the convention of openfermion's `count_qubits`, so that a qubit is
        counted even if there are no operations on it. For example,
        PauliSum("Z0+Z3").n_qubits = 4, but len(PauliSum("Z0+Z3").qubits) = 2.

        """
        return 0 if self.is_constant else max(self.qubits) + 1



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
from typing import Optional

from orquestra.quantum.circuits import Circuit
from orquestra.quantum.distributions import MeasurementOutcomeDistribution
from orquestra.quantum.measurements import Measurements
from orquestra.quantum.runners.symbolic_simulator import SymbolicSimulator


class MockCircuitRunner:

    supports_batching = False

    def __init__(self):
        super().__init__()
        self._simulator = SymbolicSimulator()

    def run_and_measure(
        self, circuit: Circuit, n_samples: int, **kwargs
    ) -> Measurements:
        return self._simulator.run_and_measure(circuit, n_samples)

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int]
    ) -> MeasurementOutcomeDistribution:
        measurements = self._simulator.run_and_measure(circuit, n_samples)
        return measurements.get_distribution()



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import io

import numpy as np
import pytest
import sympy

from orquestra.quantum.circuits import _builtin_gates, _circuit, _gates
from orquestra.quantum.circuits._serde import (
    circuit_from_dict,
    circuitset_from_dict,
    custom_gate_def_from_dict,
    deserialize_expr,
    load_circuit,
    load_circuitset,
    save_circuit,
    save_circuitset,
    serialize_expr,
    to_dict,
)

ALPHA = sympy.Symbol("alpha")
GAMMA = sympy.Symbol("gamma")
THETA = sympy.Symbol("theta")
PARAMETER_VECTOR = [sympy.Symbol("p[0]"), sympy.Symbol("p[1]")]


CUSTOM_U_GATE = _gates.CustomGateDefinition(
    "U",
    sympy.Matrix(
        [
            [THETA, GAMMA],
            [-GAMMA, THETA],
        ]
    ),
    (THETA, GAMMA),
)


# Based on ZQS-811 bug report
CUSTOM_SGD_GATE = _gates.CustomGateDefinition(
    gate_name="SGD",
    matrix=sympy.Matrix(
        [
            [1.0, 0],
            [0, -1.0j],
        ]
    ),
    params_ordering=tuple(),
)


EXAMPLE_CIRCUITS = [
    _circuit.Circuit(),
    _circuit.Circuit([_builtin_gates.X(0)]),
    _circuit.Circuit([_builtin_gates.X(2), _builtin_gates.Y(1)]),
    _circuit.Circuit(
        [
            _builtin_gates.H(0),
            _builtin_gates.CNOT(0, 1),
            _builtin_gates.RX(0)(5),
            _builtin_gates.RX(np.pi)(2),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RX(GAMMA * 2)(3),
        ]
    ),
    _circuit.Circuit(
        operations=[
            _builtin_gates.T(0),
            CUSTOM_U_GATE(1, -1)(3),
            CUSTOM_U_GATE(ALPHA, -1)(2),
        ],
    ),
    _circuit.Circuit(
        operations=[
            CUSTOM_SGD_GATE()(3),
        ],
    ),
    _circuit.Circuit(
        operations=[
            CUSTOM_U_GATE(2 + 3j, -1)(2),
        ],
    ),
    _circuit.Circuit(
        [
            _builtin_gates.H.controlled(1)(0, 1),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.Z.controlled(2)(4, 3, 0),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RY(ALPHA * GAMMA).controlled(1)(3, 2),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.X.dagger(2),
            _builtin_gates.I.dagger(4),
            _builtin_gates.Y.dagger(1),
            _builtin_gates.Z.dagger(2),
            _builtin_gates.T.dagger(7),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.Y.exp(1),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RX(-np.pi).dagger(2),
            _builtin_gates.RY(-np.pi / 2).dagger(1),
            _builtin_gates.RZ(0).dagger(0),
            _builtin_gates.PHASE(np.pi / 5).dagger(2),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RX(-np.pi).exp(2),
            _builtin_gates.RY(-np.pi / 2).exp(1),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RX(GAMMA * ALPHA).dagger(1),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.RY(PARAMETER_VECTOR[0] * PARAMETER_VECTOR[1])(1),
        ]
    ),
    _circuit.Circuit(
        [
            _builtin_gates.Z.power(0.25)(0),
            _builtin_gates.SWAP.power(0.5)(0, 1),
        ]
    ),
]


@pytest.mark.parametrize(
    "circuit",
    EXAMPLE_CIRCUITS,
)
class TestCircuitSerialization:
    def test_roundrip_results_in_same_circuit(self, circuit):
        serialized = to_dict(circuit)
        assert circuit_from_dict(serialized) == circuit

    def test_deserialized_gates_produce_matrices(self, circuit):
        deserialized_circuit = circuit_from_dict(to_dict(circuit))
        for operation in deserialized_circuit.operations:
            # matrices are computed lazily, so we have to call the getter to know if
            # we deserialized parameters properly
            operation.gate.matrix


class TestCircuitsetSerialization:
    @pytest.mark.parametrize(
        "circuitset",
        [
            [],
            list(EXAMPLE_CIRCUITS),
        ],
    )
    def test_roundrip_results_in_same_circuitset(self, circuitset):
        serialized = to_dict(circuitset)
        assert circuitset_from_dict(serialized) == circuitset


class TestCustomGateDefinitionSerialization:
    @pytest.mark.parametrize(
        "gate_def",
        [
            _gates.CustomGateDefinition(
                "V", sympy.Matrix([[THETA, GAMMA], [-GAMMA, THETA]]), (THETA, GAMMA)
            )
        ],
    )
    def test_roundtrip_gives_back_same_def(self, gate_def):
        dict_ = to_dict(gate_def)
        assert custom_gate_def_from_dict(dict_) == gate_def


class TestExpressionSerialization:
    @pytest.mark.parametrize(
        "expr,symbol_names",
        [
            (0, []),
            (1, []),
            (-1, []),
            (THETA, ["theta"]),
            (GAMMA, ["gamma"]),
            (THETA * GAMMA + 1, ["gamma", "theta"]),
            (2 + 3j, []),
            ((-1 + 2j) * THETA * GAMMA, ["gamma", "theta"]),
        ],
    )
    def test_roundtrip_results_in_equivalent_expression(self, expr, symbol_names):
        serialized = serialize_expr(expr)
        deserialized = deserialize_expr(serialized, symbol_names)
        # `deserialized == expr` wouldn't work here for complex literals because of
        # how Sympy compares expressions
        assert deserialized - expr == 0


class TestIOHelpers:
    @pytest.mark.parametrize("circuit", EXAMPLE_CIRCUITS)
    def test_load_save_circuit_roundtrip(self, circuit):
        buf = io.StringIO()
        save_circuit(circuit, buf)
        buf.seek(0)
        assert load_circuit(buf) == circuit

    def test_load_save_circuitset_roundtrip(self):
        circuitset = EXAMPLE_CIRCUITS
        buf = io.StringIO()
        save_circuitset(circuitset, buf)
        buf.seek(0)
        assert load_circuitset(buf) == circuitset



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import random
import unittest

import numpy as np

from orquestra.quantum.distributions import compute_mmd
from orquestra.quantum.distributions.target_thermal_states import (
    _get_random_ising_hamiltonian_parameters,
    convert_integer_to_ising_bitstring,
    convert_ising_bitstring_to_integer,
    get_cardinality_distribution,
    get_thermal_sampled_distribution,
    get_thermal_target_measurement_outcome_distribution,
)
from orquestra.quantum.utils import dec2bin

SEED = 14943


class TestThermalTarget(unittest.TestCase):
    def test_convert_integer_to_ising_bitstring(self):
        # Given
        number = 4
        length = 5
        expected_ising_bitstring = [-1, -1, 1, -1, -1]

        # When
        ising_bitstring = convert_integer_to_ising_bitstring(number, length)

        # Then
        self.assertListEqual(ising_bitstring, expected_ising_bitstring)

        # When
        ising_bitstring_zero = convert_integer_to_ising_bitstring(0, length)

        # Then
        self.assertListEqual(ising_bitstring_zero, [-1, -1, -1, -1, -1])

    def test_convert_ising_bitstring_to_integer(self):
        # Given
        ising_bitstring = [-1, -1, 1, -1, -1]
        expected_number = 4
        zero_bitstring = [-1, -1, -1, -1, -1]

        # When
        number = convert_ising_bitstring_to_integer(ising_bitstring)

        # Then
        self.assertEqual(number, expected_number)

        # When
        number = convert_ising_bitstring_to_integer(zero_bitstring)

        # Then
        self.assertEqual(number, 0)

    def test_ising2int2ising(self):
        # Given
        expected_number = 14
        length = 5
        zero_number = 0
        zero_length = 0
        # When
        number = convert_ising_bitstring_to_integer(
            convert_integer_to_ising_bitstring(expected_number, length)
        )

        # Then
        self.assertEqual(number, expected_number)

        # When
        number = convert_ising_bitstring_to_integer(
            convert_integer_to_ising_bitstring(zero_number, length)
        )

        # Then
        self.assertEqual(number, zero_number)

        # When
        number = convert_ising_bitstring_to_integer(
            convert_integer_to_ising_bitstring(zero_number, zero_length)
        )

        # Then
        self.assertEqual(number, zero_number)

    def test_get_random_ising_hamiltonian_parameters(self):
        # Given
        np.random.seed(seed=SEED)
        n_spins = 15

        # When
        hamiltonian_parameters = _get_random_ising_hamiltonian_parameters(n_spins)

        # Then
        self.assertEqual(
            len(hamiltonian_parameters[0]),
            n_spins,
        )
        self.assertEqual(
            hamiltonian_parameters[1].shape,
            (n_spins, n_spins),
        )

    def test_get_thermal_target_distribution_dict(self):
        # Given
        n_spins = 5
        temperature = 1.0
        expected_bitstrings = [
            tuple(dec2bin(number, n_spins)) for number in range(int(2**n_spins))
        ]
        expected_keys = expected_bitstrings[len(expected_bitstrings) :: -1]
        np.random.seed(SEED)
        external_fields = np.random.rand(n_spins)
        two_body_couplings = np.random.rand(n_spins, n_spins)
        hamiltonian_parameters = [external_fields, two_body_couplings]

        # When
        target_distribution = get_thermal_target_measurement_outcome_distribution(
            n_spins, temperature, hamiltonian_parameters
        )

        # Then
        self.assertListEqual(
            sorted(list(target_distribution.distribution_dict.keys())),
            sorted(expected_keys),
        )

        self.assertAlmostEqual(
            sum(list(target_distribution.distribution_dict.values())),
            1.0,
        )

    def test_thermal_sampled_distribution(self):
        # Given
        n_samples = 5000
        n_spins = 2
        temperature = 0.85
        np.random.seed(SEED)  # needed by our samplers
        random.seed(SEED)  # needed to make lea reproducible

        external_fields = np.random.rand(n_spins)
        two_body_couplings = np.random.rand(n_spins, n_spins)
        hamiltonian_parameters = [external_fields, two_body_couplings]
        expected_bitstrings = [
            tuple(dec2bin(number, n_spins)) for number in range(int(2**n_spins))
        ]
        expected_keys = expected_bitstrings[len(expected_bitstrings) :: -1]

        # When
        sample_distribution = get_thermal_sampled_distribution(
            n_samples, n_spins, temperature, hamiltonian_parameters
        )

        # Then
        self.assertListEqual(
            sorted(list(sample_distribution.distribution_dict.keys())),
            sorted(expected_keys),
        )

        self.assertAlmostEqual(
            sum(list(sample_distribution.distribution_dict.values())), 1
        )

    def test_samples_from_distribution(self):
        # Given
        n_samples = 10000
        n_spins = 4
        temperature = 1.0
        distance_measure_parameters = {"sigma": 1.0}
        np.random.seed(SEED)  # needed by our samplers
        random.seed(SEED)  # needed to make lea reproducible

        external_fields = np.random.rand(n_spins)
        two_body_couplings = np.random.rand(n_spins, n_spins)
        hamiltonian_parameters = [external_fields, two_body_couplings]

        actual = get_thermal_target_measurement_outcome_distribution(
            n_spins, temperature, hamiltonian_parameters
        )
        model = get_thermal_sampled_distribution(
            n_samples, n_spins, temperature, hamiltonian_parameters
        )

        # When
        mmd = compute_mmd(actual, model, distance_measure_parameters)

        # Then
        self.assertLess(mmd, 1e-4)

    def test_cardinality_distribution(self):
        # Given
        n_samples = 10000
        n_spins = 4
        temperature = 1.0
        np.random.seed(SEED)  # needed by our samplers
        random.seed(SEED)  # needed to make lea reproducible

        external_fields = np.random.rand(n_spins)
        two_body_couplings = np.random.rand(n_spins, n_spins)
        hamiltonian_parameters = [external_fields, two_body_couplings]
        sampled_distribution = get_thermal_sampled_distribution(
            n_samples, n_spins, temperature, hamiltonian_parameters
        )

        # When
        cardinality_distr = get_cardinality_distribution(
            n_samples, n_spins, sampled_distribution
        )

        # Then
        assert all([spin_cardinality >= 0 for spin_cardinality in cardinality_distr])


if __name__ == "__main__":
    unittest.main()



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import os

from orquestra.quantum.circuits.layouts import (
    CircuitConnectivity,
    CircuitLayers,
    build_circuit_layers_and_connectivity,
    load_circuit_connectivity,
    load_circuit_layers,
    load_circuit_ordering,
    save_circuit_connectivity,
    save_circuit_layers,
    save_circuit_ordering,
)


class TestCircuitLayers:
    def test_dict_io(self):
        # Given
        layers = CircuitLayers([[(0, 1), (2, 3)], [(1, 2), (3, 0)]])

        # When
        layers_dict = layers.to_dict()
        recreated_layers = CircuitLayers.from_dict(layers_dict)

        # Then
        assert len(layers.layers) == len(recreated_layers.layers)
        for layer, recreated_layer in zip(layers.layers, recreated_layers.layers):
            assert layer == recreated_layer

    def test_circuit_layers_io(self):
        # Given
        layers = CircuitLayers([[(0, 1), (2, 3)], [(1, 2), (3, 0)]])

        # When
        save_circuit_layers(layers, "layers.json")
        loaded_layers = load_circuit_layers("layers.json")

        # Then
        assert len(layers.layers) == len(loaded_layers.layers)
        for layer, loaded_layer in zip(layers.layers, loaded_layers.layers):
            assert layer == loaded_layer
        os.remove("layers.json")

    def test_circuit_ordering_io(self):
        # Given
        ordering = [0, 1, 3, 2]

        # When
        save_circuit_ordering(ordering, "ordering.json")
        loaded_ordering = load_circuit_ordering("ordering.json")

        # Then
        assert loaded_ordering == ordering
        os.remove("ordering.json")

    def test_circuit_connectivity_io(self):
        # Given
        connectivity = CircuitConnectivity([(0, 1), (1, 2), (2, 3), (3, 0)])
        # When
        save_circuit_connectivity(connectivity, "connectivity.json")
        loaded_connectivity = load_circuit_connectivity("connectivity.json")
        # Then
        assert len(connectivity.connectivity) == len(loaded_connectivity.connectivity)
        for connection, loaded_connection in zip(
            connectivity.connectivity, loaded_connectivity.connectivity
        ):
            assert connection == loaded_connection
        os.remove("connectivity.json")

    def test_build_circuit_layers_and_connectivity(self):
        # Sycamore
        # Given
        test_layers = [
            [(3, 1), (4, 7)],
            [(3, 6), (4, 2), (5, 8)],
            [(3, 1), (4, 2)],
            [(3, 6), (4, 7), (5, 8)],
            [(3, 0), (4, 8), (5, 2)],
            [(3, 7), (4, 1)],
            [(3, 0), (4, 1), (5, 2)],
            [(3, 7), (4, 8)],
        ]
        test_connectivity = [
            (3, 0),
            (3, 1),
            (4, 1),
            (4, 2),
            (5, 2),
            (3, 6),
            (3, 7),
            (4, 7),
            (4, 8),
            (5, 8),
        ]
        # When
        connectivity, layers = build_circuit_layers_and_connectivity(3, 3, "sycamore")

        # Then
        for layer, test_layer in zip(layers.layers, test_layers):
            assert layer == test_layer

        for row, test_row in zip(connectivity.connectivity, test_connectivity):
            assert row == test_row

        # Nearest-Neighbors
        # Given
        test_layers = [[(0, 1), (2, 3)], [(1, 2)]]
        test_connectivity = [(0, 1), (2, 3), (1, 2)]

        # When
        connectivity, layers = build_circuit_layers_and_connectivity(
            4, "nearest-neighbor"
        )

        # Then
        for layer, test_layer in zip(layers.layers, test_layers):
            assert layer == test_layer

        for row, test_row in zip(connectivity.connectivity, test_connectivity):
            assert row == test_row



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
from typing import Dict, List, Optional, Sequence, Union

from ..api.circuit_runner import BaseCircuitRunner, CircuitRunner
from ..circuits import Circuit, to_dict
from ..distributions import MeasurementOutcomeDistribution
from ..measurements import Measurements


class MeasurementTrackingBackend(BaseCircuitRunner):
    """A wrapper class for a runner that tracks all measurements. The measurements
    are stored in the raw_circuit_data variable as a list of measurement objects.
    """

    def __init__(
        self,
        inner_backend: CircuitRunner,
        raw_data_file_name: str,
        record_bitstrings: Optional[bool] = False,
    ):
        """Create a wrapper runner around inner_backend that keeps track of all the
        measurement data collected during an experiment.

        Args:
            inner_backend (QuantumBackend): Backend for which measurements are recorded
            record_bitstrings (bool, optional): Record every measured bitstring. May
                lead to large amounts of stored data. Defaults to False.
        """
        super().__init__()
        self.record_bitstrings: Optional[bool] = record_bitstrings
        self.inner_backend: CircuitRunner = inner_backend
        self.raw_data: List[Dict] = []
        self.type: str = inner_backend.__class__.__name__
        self.raw_data_file_name = raw_data_file_name

    def _run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """Method for executing the circuit and measuring the outcome.

        Args:
            circuit: quantum circuit to be executed.
            n_samples: The number of samples to collect.
        """
        measurement = self.inner_backend.run_and_measure(circuit, n_samples)
        self.record_raw_measurement_data(circuit, measurement)
        self.save_raw_data()
        return measurement

    def run_batch_and_measure(
        self, circuits: Sequence[Circuit], n_samples: Union[int, Sequence[int]]
    ) -> List[Measurements]:
        """Run a set of circuits and measure a certain number of bitstrings.

        Args:
            circuits: The circuits to execute.
            n_samples: The number of samples to collect for each circuit.
        """
        self._n_circuits_executed += len(circuits)
        self._n_jobs_executed += 1
        measurements = self.inner_backend.run_batch_and_measure(circuits, n_samples)
        for circuit, measurement in zip(circuits, measurements):
            self.record_raw_measurement_data(circuit, measurement)
        self.save_raw_data()
        return measurements

    def record_raw_measurement_data(
        self, circuit: Circuit, measurement: Measurements
    ) -> None:
        """Append data from a measurement to self.raw_data.

        Args:
            circuit: Implemented circuit.
            measurement: Implemented measurement.
        """
        raw_data_dict = {
            "data_type": "measurement",
            "device": self.type,
            "circuit": to_dict(circuit),
            "counts": measurement.get_counts(),
            "number_of_gates": len(circuit.operations),
            "number_of_shots": len(measurement.bitstrings),
        }
        if self.record_bitstrings:
            raw_data_dict["bitstrings"] = [
                list(map(int, list(bitstring))) for bitstring in measurement.bitstrings
            ]
        self.raw_data.append(raw_data_dict)

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int]
    ) -> MeasurementOutcomeDistribution:
        """Calculates a measurement outcome distribution.

        Args:
            circuit: quantum circuit to be executed.

        Returns:
            Probability distribution of getting specific bistrings.

        """
        distribution = self.inner_backend.get_measurement_outcome_distribution(
            circuit, n_samples
        )
        self.raw_data.append(
            {
                "data_type": "measurement outcome distribution",
                "device": self.type,
                "circuit": to_dict(circuit),
                "distribution": repr(distribution),
                "number_of_gates": len(circuit.operations),
                "number_of_shots": n_samples,
            }
        )
        self.save_raw_data()
        return distribution

    def save_raw_data(self) -> None:
        with open(self.raw_data_file_name, "w+") as f:
            data = {
                "raw-data": self.raw_data,
            }
            f.write(json.dumps(data))
        self.raw_data = []



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
### The following code is derived from:
### https://github.com/GiggleLiu/QuantumCircuitBornMachine/blob/master/LICENSE
### The original resource is under MIT license, which is pasted below for convenience

# MIT License
# Copyright (c) 2017 Leo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, Dict, Sequence

import numpy as np

if TYPE_CHECKING:
    from ..distributions import MeasurementOutcomeDistribution


def compute_rbf_kernel(x_i: np.ndarray, y_j: np.ndarray, sigma: float) -> np.ndarray:
    """Compute the gaussian (RBF) kernel matrix.

    The computed kernel matrix K has elements :math:`K_ij = exp(-gamma |x_i - y_j|^2)`
    where :math:`gamma = 1/(2*sigma)`.

    Args:
        x_i: Samples A (integers).
        y_j: Samples B (integers).
        sigma: The bandwidth of the gaussian kernel.

    Returns:
        np.ndarray: The gaussian kernel matrix.
    """
    exponent = np.abs(x_i[:, None] - y_j[None, :]) ** 2
    try:
        gamma = 1.0 / (2 * sigma)
    except ZeroDivisionError as error:
        print("Handling run-time error:", error)
        raise
    kernel_matrix = np.exp(-gamma * exponent)
    return kernel_matrix


def compute_multi_rbf_kernel(
    x_i: np.ndarray, y_j: np.ndarray, sigmas: Sequence[float]
) -> np.ndarray:
    """Compute the multi-gaussian (RBF) kernel matrix K, with

    :math:`K_ij = 1/N * Sum_n [exp(-gamma_n |x_i - y_j|^2)]`

    with n = 1,...,N and gamma = 1/(2*sigma).

    Args:
        x_i: Samples A (integers).
        y_j: Samples B (integers).
        sigmas: The list of bandwidths of the multi-gaussian kernel.

    Returns:
        np.ndarray: The gaussian kernel matrix.
    """
    exponent = np.abs(x_i[:, None] - y_j[None, :]) ** 2
    kernel_matrix = np.zeros(exponent.shape)
    for sigma in sigmas:
        try:
            gamma = 1.0 / (2 * sigma)
        except ZeroDivisionError as error:
            print("Handling run-time error:", error)
            raise
        kernel_matrix += np.exp(-gamma * exponent)
    return kernel_matrix / len(sigmas)


def compute_mmd(
    target_distribution: "MeasurementOutcomeDistribution",
    measured_distribution: "MeasurementOutcomeDistribution",
    distance_measure_parameters: Dict,
) -> float:
    """Compute the squared Maximum Mean Discrepancy (MMD) distance measure between
    between a target distribution and a measured distribution.
    Reference: arXiv.1804.04168.

        Args:
            target_distribution: The target probability distribution.
            measured_distribution: The measured probability distribution.
            distance_measure_parameters: dictionary with the following keys:

                - sigma (float/np.array): the bandwidth parameter used to compute the
                  single/multi gaussian kernel. The default value is 1.0.

        Returns:
            The value of the maximum mean discrepancy.
    """

    sigma = distance_measure_parameters.get("sigma", 1.0)
    target_keys = target_distribution.distribution_dict.keys()
    measured_keys = measured_distribution.distribution_dict.keys()
    all_keys = set(target_keys).union(measured_keys)

    target_values = []
    measured_values = []
    for bitstring in all_keys:
        # Add 0 to the values list whenever a bistrings isn't found among the keys.
        target_values.append(target_distribution.distribution_dict.get(bitstring, 0))
        measured_values.append(
            measured_distribution.distribution_dict.get(bitstring, 0)
        )

    basis = np.asarray(
        [int("".join(map(str, item)), 2) for item in all_keys]
    )  # Digit Tuple to int
    if not hasattr(sigma, "__len__"):
        kernel_matrix = compute_rbf_kernel(basis, basis, sigma)
    else:
        kernel_matrix = compute_multi_rbf_kernel(basis, basis, sigma)

    diff = np.array(target_values) - np.array(measured_values)
    return diff.dot(kernel_matrix.dot(diff))



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
"""This module contains functions and classes for pauli representations of operators."""
from ._io import (
    convert_dict_to_op,
    convert_op_to_dict,
    get_pauli_strings,
    load_operator,
    load_operator_set,
    save_operator,
    save_operator_set,
)
from ._openfermion_utils import (
    EQ_TOLERANCE,
    expectation,
    get_sparse_operator,
    hermitian_conjugated,
    is_hermitian,
)
from ._pauli_operators import PauliRepresentation, PauliSum, PauliTerm
from ._utils import get_expectation_value, reverse_qubit_order



#!/usr/bin/env python
################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################

import site
import sys

from pygit2 import Repository
from setuptools import find_namespace_packages, setup

try:
    from subtrees.z_quantum_actions.setup_extras import extras
except ImportError:
    print("Unable to import extras", file=sys.stderr)
    extras = {"dev": []}

# Workaound for https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# This next stanza will make sure that if we're working on the @dev branch of
# this repo, that we use the @dev branch fo orquestra-sdk.
branch_name = Repository(".").head.shorthand
orquestra_sdk_dep = "orquestra-sdk @ git+ssh://git@github.com/zapatacomputing/orquestra-sdk.git"  # noqa: E501
if "feature" in branch_name:
    orquestra_sdk_dep += "@dev"

setup(
    name="orquestra-bogus",
    description="Monitoring Orquestra workflows.",
    package_dir={"": "src/python"},
    packages=find_namespace_packages(include=["orquestra.*"], where="src/python"),
    include_package_data=True,
    license="LICENSE",
    install_requires=[
        "pydantic",
        "ray[default]",
        orquestra_sdk_dep,
    ],
    extras_require=extras,
    # Without this, users of this library would get mypy errors. See also:
    # https://github.com/python/mypy/issues/7508#issuecomment-531965557
    zip_safe=False,
)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from functools import partial
from itertools import cycle

import numpy as np
import pytest
import sympy

from orquestra.quantum.api.circuit_runner import BaseCircuitRunner
from orquestra.quantum.api.estimation import EstimationTask
from orquestra.quantum.circuits import RX, RY, RZ, Circuit, H, X
from orquestra.quantum.estimation import (
    calculate_exact_expectation_values,
    estimate_expectation_values_by_averaging,
    evaluate_estimation_circuits,
    evaluate_non_measured_estimation_tasks,
    split_estimation_tasks_to_measure,
)
from orquestra.quantum.measurements import ExpectationValues, Measurements
from orquestra.quantum.operators import PauliSum, PauliTerm
from orquestra.quantum.runners.symbolic_simulator import SymbolicSimulator
from orquestra.quantum.testing import MockCircuitRunner


class TestEstimatorUtils:
    @pytest.fixture()
    def frame_operators(self):
        operators = [
            PauliTerm("2*Z1*Z2"),
            PauliTerm("2*Z0*Z3"),
            PauliTerm("Z2", -1),
        ]

        return operators

    @pytest.fixture()
    def circuits(self):
        circuits = [Circuit() for _ in range(5)]

        circuits[1] += RX(1.2)(0)
        circuits[1] += RY(1.5)(1)
        circuits[1] += RX(-0.0002)(0)
        circuits[1] += RY(0)(1)

        for circuit in circuits[2:]:
            circuit += RX(sympy.Symbol("theta_0"))(0)
            circuit += RY(sympy.Symbol("theta_1"))(1)
            circuit += RX(sympy.Symbol("theta_2"))(0)
            circuit += RY(sympy.Symbol("theta_3"))(1)

        return circuits

    def test_evaluate_estimation_circuits_all_symbols(
        self,
        circuits,
    ):
        symbols_maps = [
            [
                (sympy.Symbol("theta_0"), 0),
                (sympy.Symbol("theta_1"), 0),
                (sympy.Symbol("theta_2"), 0),
                (sympy.Symbol("theta_3"), 0),
            ]
            for _ in circuits
        ]
        evaluate_circuits = partial(
            evaluate_estimation_circuits,
            symbols_maps=symbols_maps,
        )
        operator = PauliSum()
        estimation_tasks = [
            EstimationTask(operator, circuit, 1) for circuit in circuits
        ]

        new_estimation_tasks = evaluate_circuits(estimation_tasks)

        for new_task in new_estimation_tasks:
            assert len(new_task.circuit.free_symbols) == 0

    @pytest.mark.parametrize(
        ",".join(
            [
                "estimation_tasks",
                "ref_estimation_tasks_to_measure",
                "ref_non_measured_estimation_tasks",
                "ref_indices_to_measure",
                "ref_non_measured_indices",
            ]
        ),
        [
            (
                [
                    EstimationTask(
                        PauliSum([PauliTerm("Z0", 2), PauliTerm("3*Z1*Z2")]),
                        Circuit([X(0)]),
                        10,
                    ),
                    EstimationTask(
                        PauliSum(
                            [
                                PauliTerm("Z0", 2),
                                PauliTerm("3*Z1*Z2"),
                                PauliTerm("I0", 4),
                            ]
                        ),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [
                    EstimationTask(
                        PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([X(0)]),
                        10,
                    ),
                    EstimationTask(
                        PauliTerm("I0", 4) + PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [],
                [0, 1, 2],
                [],
            ),
            (
                [
                    EstimationTask(
                        PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([X(0)]),
                        10,
                    ),
                    EstimationTask(
                        PauliTerm("I0", 4),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [
                    EstimationTask(
                        PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([X(0)]),
                        10,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [EstimationTask(PauliTerm("I0", 4), Circuit([RZ(np.pi / 2)(0)]), 1000)],
                [0, 2],
                [1],
            ),
            (
                [
                    EstimationTask(PauliTerm("I0", -3), Circuit([X(0)]), 0),
                    EstimationTask(
                        PauliTerm("I0", 4) + PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [
                    EstimationTask(
                        PauliTerm("I0", 4) + PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        17,
                    ),
                ],
                [
                    EstimationTask(PauliTerm("I0", -3), Circuit([X(0)]), 0),
                ],
                [1, 2],
                [0],
            ),
            (
                [
                    EstimationTask(PauliTerm("I0", -3), Circuit([X(0)]), 0),
                    EstimationTask(
                        PauliTerm("I0", 4) + PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        0,
                    ),
                ],
                [
                    EstimationTask(
                        PauliTerm("I0", 4) + PauliTerm("Z0", 2) + PauliTerm("3*Z1*Z2"),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                ],
                [
                    EstimationTask(PauliTerm("I0", -3), Circuit([X(0)]), 0),
                    EstimationTask(
                        PauliTerm("Z3", 4),
                        Circuit([RY(np.pi / 2)(0)]),
                        0,
                    ),
                ],
                [1],
                [0, 2],
            ),
        ],
    )
    def test_split_estimation_tasks_to_measure(
        self,
        estimation_tasks,
        ref_estimation_tasks_to_measure,
        ref_non_measured_estimation_tasks,
        ref_indices_to_measure,
        ref_non_measured_indices,
    ):

        (
            estimation_task_to_measure,
            non_measured_estimation_tasks,
            indices_to_measure,
            indices_for_non_measureds,
        ) = split_estimation_tasks_to_measure(estimation_tasks)

        assert estimation_task_to_measure == ref_estimation_tasks_to_measure
        assert non_measured_estimation_tasks == ref_non_measured_estimation_tasks
        assert indices_to_measure == ref_indices_to_measure
        assert ref_non_measured_indices == indices_for_non_measureds

    @pytest.mark.parametrize(
        "estimation_tasks,ref_expectation_values",
        [
            (
                [
                    EstimationTask(
                        PauliTerm("I0", 4),
                        Circuit([RZ(np.pi / 2)(0)]),
                        1000,
                    ),
                ],
                [
                    ExpectationValues(
                        np.asarray([4.0]),
                        correlations=[np.asarray([[0.0]])],
                        estimator_covariances=[np.asarray([[0.0]])],
                    ),
                ],
            ),
            (
                [
                    EstimationTask(
                        PauliTerm("I0", -0.5) + PauliTerm("I0", -2.5),
                        Circuit([X(0)]),
                        0,
                    ),
                    EstimationTask(
                        PauliTerm("I0", 0.001), Circuit([RZ(np.pi / 2)(0)]), 2
                    ),
                    EstimationTask(
                        PauliTerm("Z1", 2.5) + PauliTerm("1*Z2*Z3"),
                        Circuit([RY(np.pi / 2)(0)]),
                        0,
                    ),
                ],
                [
                    ExpectationValues(
                        np.asarray([-3.0]),
                        correlations=[np.asarray([[0.0]])],
                        estimator_covariances=[np.asarray([[0.0]])],
                    ),
                    ExpectationValues(
                        np.asarray([0.001]),
                        correlations=[np.asarray([[0.0]])],
                        estimator_covariances=[np.asarray([[0.0]])],
                    ),
                    ExpectationValues(
                        np.asarray([0.0]),
                        correlations=[np.asarray([[0.0]])],
                        estimator_covariances=[np.asarray([[0.0]])],
                    ),
                ],
            ),
        ],
    )
    def test_evaluate_non_measured_estimation_tasks(
        self, estimation_tasks, ref_expectation_values
    ):

        expectation_values = evaluate_non_measured_estimation_tasks(estimation_tasks)

        for ex_val, ref_ex_val in zip(expectation_values, ref_expectation_values):
            assert np.allclose(ex_val.values, ref_ex_val.values)
            assert np.allclose(ex_val.correlations, ref_ex_val.correlations)
            assert np.allclose(
                ex_val.estimator_covariances, ref_ex_val.estimator_covariances
            )

    @pytest.mark.parametrize(
        "estimation_tasks",
        [
            (
                [
                    EstimationTask(
                        PauliSum([PauliTerm("I0", -2.5), PauliTerm("Z1", -0.5)]),
                        Circuit([X(0)]),
                        1,
                    ),
                ]
            ),
            (
                [
                    EstimationTask(
                        PauliTerm("Z0", 0.001),
                        Circuit([RZ(np.pi / 2)(0)]),
                        0,
                    ),
                    EstimationTask(
                        PauliTerm("I0", 2.0), Circuit([RZ(np.pi / 2)(0)]), 2
                    ),
                    EstimationTask(
                        PauliTerm("1.5*Z0*Z1"),
                        Circuit([RY(np.pi / 2)(0)]),
                        10,
                    ),
                ]
            ),
        ],
    )
    def test_evaluate_non_measured_estimation_tasks_fails_with_non_zero_shots(
        self, estimation_tasks
    ):
        with pytest.raises(RuntimeError):
            _ = evaluate_non_measured_estimation_tasks(estimation_tasks)


TEST_CASES_EIGENSTATES = [
    (
        [
            EstimationTask(
                PauliTerm("Z0"), circuit=Circuit([X(0)]), number_of_shots=10
            ),
            EstimationTask(
                PauliTerm("I0", coefficient=2.0),
                circuit=Circuit([RY(np.pi / 4)(0)]),
                number_of_shots=30,
            ),
            # tests correlation
            EstimationTask(
                PauliSum([PauliTerm("Z0"), PauliTerm("Z1")]),
                circuit=Circuit([X(0), X(1)]),
                number_of_shots=10,
            ),
            # tests negative correlation
            EstimationTask(
                PauliSum([PauliTerm("Z0"), PauliTerm("Z1")]),
                circuit=Circuit([X(1)]),
                number_of_shots=10,
            ),
        ],
        [
            ExpectationValues(np.array([-1]), [np.array([[1]])], [np.array([[0]])]),
            ExpectationValues(np.array([2]), [np.array([[0]])], [np.array([[0]])]),
            ExpectationValues(
                np.array([-1, -1]),
                [np.array([[1, 1], [1, 1]])],
                [np.array([[0, 0], [0, 0]])],
            ),
            ExpectationValues(
                np.array([1, -1]),
                [np.array([[1, -1], [-1, 1]])],
                [np.array([[0, 0], [0, 0]])],
            ),
        ],
    ),
]
TEST_CASES_NONEIGENSTATES = [
    (
        [
            EstimationTask(
                PauliTerm("Z0"),
                circuit=Circuit([H(0)]),
                number_of_shots=10000,
            ),
            EstimationTask(
                PauliTerm("Z0", coefficient=-2),
                circuit=Circuit([RY(np.pi / 4)(0)]),
                number_of_shots=10000,
            ),
        ],
        [
            ExpectationValues(np.array([0]), [np.array([[1]])], [np.array([[0.001]])]),
            ExpectationValues(
                np.array([-2 * (np.cos(np.pi / 8) ** 2 - 0.5) * 2]),
                [np.array([[4]])],
                [np.array([[0.002]])],
            ),
        ],
    ),
]
""" When number_of_shots is high, the covariance will tend to zero. So we need extra
test cases to ensure covariance is being caluclated correctly.

Since we generate the data with MockBackendForTestingCovariancewhenNumberOfShotsIsLow
the estimation tasks can be arbitrary. They just need a valid circuit, number of shots
and have 2 terms in the ising operator."""
TEST_CASES_NONEIGENSTATES_WITH_LOW_NUMBER_OF_SHOTS = [
    (
        4
        * [
            EstimationTask(
                PauliSum([PauliTerm("Z0"), PauliTerm("Z1")]),
                circuit=Circuit([X(0), X(1)]),
                number_of_shots=10,
            ),
        ],
        [
            ExpectationValues(
                np.array([0]),
                [np.array([[1, 0], [0, 1]])],
                [np.array([[0, 0], [0, 0.5]])],
            ),
            ExpectationValues(
                np.array([0]),
                [np.array([[1, 0], [0, 1]])],
                [np.array([[0.5, 0], [0, 0]])],
            ),
            ExpectationValues(
                np.array([0]),
                [np.array([[1, -1], [-1, 1]])],
                [np.array([[0.5, -0.5], [-0.5, 0.5]])],
            ),
            ExpectationValues(
                np.array([0]),
                [np.array([[1, 1], [1, 1]])],
                [np.array([[0.05, 0.05], [0.05, 0.05]])],
            ),
        ],
    ),
]


# needs its own class otherwise issues arise with calling run_circuitset_and_measure.
class MockBackendForTestingCovariancewhenNumberOfShotsIsLow(BaseCircuitRunner):
    def __init__(self):
        super().__init__()
        self._measurements = iter(
            cycle(
                [
                    Measurements([(0, 1), (0, 0)]),
                    Measurements([(1, 0), (0, 0)]),
                    Measurements([(1, 0), (0, 1)]),
                    Measurements([(1, 1), (0, 0)] * 10),
                ]
            )
        )

    def _run_and_measure(self, circuit, shots_per_circuit):
        return next(self._measurements)


class TestBasicEstimationMethods:
    @pytest.fixture()
    def simulator(self):
        return SymbolicSimulator()

    @pytest.fixture()
    def estimation_tasks(self):
        task_1 = EstimationTask(
            PauliTerm("Z0"), circuit=Circuit([X(0)]), number_of_shots=10
        )
        task_2 = EstimationTask(
            PauliTerm("Z0"),
            circuit=Circuit([RY(np.pi / 2)(0)]),
            number_of_shots=20,
        )
        task_3 = EstimationTask(
            PauliTerm("I0", coefficient=2.0),
            circuit=Circuit([RY(np.pi / 4)(0)]),
            number_of_shots=30,
        )
        return [task_1, task_2, task_3]

    @pytest.fixture()
    def target_expectation_values(self):
        return [ExpectationValues(-1), ExpectationValues(0), ExpectationValues(2)]

    @pytest.mark.parametrize(
        "estimation_tasks,target_expectations", TEST_CASES_EIGENSTATES
    )
    def test_estimate_expectation_values_by_averaging_for_eigenstates(
        self, simulator, estimation_tasks, target_expectations
    ):
        expectation_values_list = estimate_expectation_values_by_averaging(
            simulator, estimation_tasks
        )
        for expectation_values, target, task in zip(
            expectation_values_list, target_expectations, estimation_tasks
        ):
            assert len(expectation_values.values) == len(task.operator.terms)
            np.testing.assert_array_equal(expectation_values.values, target.values)

    @pytest.mark.parametrize(
        "estimation_tasks,target_expectations", TEST_CASES_NONEIGENSTATES
    )
    def test_estimate_expectation_values_by_averaging_for_non_eigenstates(
        self, simulator, estimation_tasks, target_expectations
    ):

        expectation_values_list = estimate_expectation_values_by_averaging(
            simulator, estimation_tasks
        )
        for expectation_values, target, task in zip(
            expectation_values_list, target_expectations, estimation_tasks
        ):
            assert len(expectation_values.values) == len(task.operator.terms)
            np.testing.assert_allclose(
                expectation_values.values, target.values, atol=0.1
            )

    @pytest.mark.skip(reason="Temporary disabled")
    @pytest.mark.parametrize(
        "estimation_tasks,target_expectations",
        TEST_CASES_EIGENSTATES + TEST_CASES_NONEIGENSTATES,
    )
    def test_calculate_exact_expectation_values(
        self, simulator, estimation_tasks, target_expectations
    ):
        expectation_values_list = calculate_exact_expectation_values(
            simulator, estimation_tasks
        )
        for expectation_values, target, task in zip(
            expectation_values_list, target_expectations, estimation_tasks
        ):
            assert len(expectation_values.values) == len(task.operator.terms)
            np.testing.assert_array_almost_equal(
                expectation_values.values, target.values
            )

    @pytest.mark.parametrize(
        "estimation_tasks,target_expectations", TEST_CASES_EIGENSTATES
    )
    def test_covariance_and_correlations_when_averaging_for_eigenstates(
        self, simulator, estimation_tasks, target_expectations
    ):
        expectation_values_list = estimate_expectation_values_by_averaging(
            simulator, estimation_tasks
        )
        for expectation_values, target, task in zip(
            expectation_values_list, target_expectations, estimation_tasks
        ):
            for cov, corr, target_corr, target_cov in zip(
                expectation_values.estimator_covariances,
                expectation_values.correlations,
                target.correlations,
                target.estimator_covariances,
            ):
                assert len(corr) == len(task.operator.terms)
                assert len(cov) == len(task.operator.terms)

                np.testing.assert_array_almost_equal(corr, target_corr, decimal=2)
                np.testing.assert_array_almost_equal(cov, target_cov, decimal=2)

    @pytest.mark.parametrize(
        "estimation_tasks,target_expectations",
        TEST_CASES_NONEIGENSTATES,
    )
    def test_correlation_and_covariance_when_averaging_for_non_eigenstates(
        self, simulator, estimation_tasks, target_expectations
    ):

        expectation_values_list = estimate_expectation_values_by_averaging(
            simulator, estimation_tasks
        )
        for expectation_values, target, task in zip(
            expectation_values_list, target_expectations, estimation_tasks
        ):
            for cov, corr, target_corr, target_cov in zip(
                expectation_values.estimator_covariances,
                expectation_values.correlations,
                target.correlations,
                target.estimator_covariances,
            ):
                assert len(corr) == len(task.operator.terms)
                assert len(cov) == len(task.operator.terms)

                np.testing.assert_array_almost_equal(corr, target_corr, decimal=1)
                # All covariances should be close to 0 when number_of_shots is high.
                np.testing.assert_array_almost_equal(target_cov, cov, decimal=2)

    @pytest.mark.parametrize(
        "mock_estimation_tasks, target_expectations",
        TEST_CASES_NONEIGENSTATES_WITH_LOW_NUMBER_OF_SHOTS,
    )
    def test_covariance_when_averaging_for_non_eigenstates_and_number_of_shots_is_low(
        self, mock_estimation_tasks, target_expectations
    ):

        expectation_values_list = estimate_expectation_values_by_averaging(
            MockBackendForTestingCovariancewhenNumberOfShotsIsLow(),
            mock_estimation_tasks,
        )
        for expectation_values, target in zip(
            expectation_values_list, target_expectations
        ):
            np.testing.assert_array_almost_equal(
                expectation_values.correlations,
                target.correlations,
                decimal=2,
            )
            np.testing.assert_array_almost_equal(
                expectation_values.estimator_covariances,
                target.estimator_covariances,
                decimal=2,
            )

    def test_calculate_exact_expectation_values_fails_with_non_simulator(
        self, estimation_tasks
    ):
        backend = MockCircuitRunner()
        with pytest.raises(AttributeError):
            _ = calculate_exact_expectation_values(backend, estimation_tasks)

    def test_no_measured_estimation_tasks_provided(simulator):
        estimation_tasks = [
            EstimationTask(PauliTerm("Z0"), Circuit([H(0)]), 0),
        ]

        target = [
            ExpectationValues(np.array([0.0]), [np.array([[0.0]])], [np.array([0.0])])
        ]

        expectation_values_list = estimate_expectation_values_by_averaging(
            simulator, estimation_tasks
        )

        assert len(expectation_values_list) == 1
        assert expectation_values_list[0] == target[0]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from dataclasses import dataclass
from typing import List, Optional, Protocol

import numpy as np

from ..circuits import Circuit
from ..measurements import ExpectationValues
from ..operators import PauliRepresentation
from .circuit_runner import CircuitRunner


@dataclass(frozen=True)
class EstimationTask:
    """
    Data class defining an estimation problem.

    Args:
        operator: Operator for which we want to calculate the expectation values
        circuit: Circuit used for evaluating the operator
        constraints: Define constraints used in the estimation process,
            e.g. number of shots or target accuracy.
    """

    operator: PauliRepresentation
    circuit: Circuit
    number_of_shots: Optional[int]


class EstimationPreprocessor(Protocol):
    """Protocol defining function which transforms a list of EstimationTasks
    into another list of EstimationTasks.
    """

    def __call__(self, estimation_tasks: List[EstimationTask]) -> List[EstimationTask]:
        ...


class EstimateExpectationValues(Protocol):
    """Protocol defining a function that estimates expectation values for a list of
    estimation tasks. Implementations of this protocol should obey the following rules:

    1. Return one ExpectationValue for each EstimationTask.
    2. The order in which ExpectationValues are returned should match the order
        in which EstimationTasks were provided.
    3. Number of entries in each ExpectationValue is not restricted.
    4. Output ExpectationValue should include coefficients of the terms/operators.
    5. estimation_tasks can include tasks where operator consists of a constant term or
        contains a constant term. The implemented method should include the
        contributions of such constant terms in the return value.
    """

    def __call__(
        self, runner: CircuitRunner, estimation_tasks: List[EstimationTask]
    ) -> List[ExpectationValues]:
        """Estimate expectation values using given runner."""


class EstimationTasksFactory(Protocol):
    """Factory from producing estimation tasks from R^n vectors.

    For instance, this can be used with ansatzes where produced estimation tasks
    are evaluating circuit.
    """

    def __call__(self, parameters: np.ndarray) -> List[EstimationTask]:
        """Produce estimation tasks for given parameters."""



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for operator_utils.py"""
import unittest

import numpy
from scipy.sparse import csc_matrix

from orquestra.quantum.operators import (
    PauliSum,
    PauliTerm,
    hermitian_conjugated,
    is_hermitian,
)


class HermitianConjugatedTest(unittest.TestCase):
    def test_hermitian_conjugated_pauli_term(self):
        """Test conjugating PauliTerms."""
        op = PauliTerm("X0*Y1", 2.0)
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertEqual(op_hc, correct_op)

        op = PauliTerm("X0*Y1", 2.0j)
        op_hc = hermitian_conjugated(op)
        correct_op = PauliTerm("X0*Y1", -2.0j)
        self.assertEqual(op_hc, correct_op)

    def test_hermitian_conjugated_pauli_sum(self):
        """Test conjugating PauliSums."""
        op = PauliTerm("X0*Y1", 2.0) + PauliTerm("Z4*X5*Y7", 3.0j)
        op_hc = hermitian_conjugated(op)
        correct_op = PauliTerm("X0*Y1", 2.0) + PauliTerm("Z4*X5*Y7", -3.0j)
        self.assertEqual(op_hc, correct_op)

    def test_hermitian_conjugate_empty(self):
        op = PauliSum()
        op = hermitian_conjugated(op)
        self.assertEqual(op, PauliSum())

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = hermitian_conjugated(1)


class IsHermitianTest(unittest.TestCase):
    def test_pauli_operator_zero(self):
        op = PauliSum()
        self.assertTrue(is_hermitian(op))

    def test_pauli_operator_identity(self):
        op = PauliTerm.identity()
        self.assertTrue(is_hermitian(op))

    def test_pauli_operator_nonhermitian(self):
        op = PauliTerm("X0*Y2*Z5", 1.0 + 2.0j)
        self.assertFalse(is_hermitian(op))

    def test_pauli_operator_hermitian(self):
        op = PauliTerm("X0*Y2*Z5", 1.0 + 2.0j)
        op += PauliTerm("X0*Y2*Z5", 1.0 - 2.0j)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_zero(self):
        op = numpy.zeros((4, 4))
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_identity(self):
        op = numpy.eye(4)
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_nonhermitian(self):
        op = numpy.arange(16).reshape((4, 4))
        self.assertFalse(is_hermitian(op))
        op = csc_matrix(op)
        self.assertFalse(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_hermitian(self):
        op = numpy.arange(16, dtype=complex).reshape((4, 4))
        op += 1.0j * op
        op += op.T.conj()
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = is_hermitian("a")



import pytest

from orquestra.quantum.circuits import CNOT, H, X
from orquestra.quantum.circuits._circuit import Circuit
from orquestra.quantum.circuits._itertools import (
    combine_bitstrings,
    combine_measurement_counts,
    expand_sample_sizes,
    split_into_batches,
)
from orquestra.quantum.measurements import Measurements


class TestBatchingWithInvalidInputs:
    @pytest.mark.parametrize("n_samples_per_circuit", [[100], [50, 100, 50]])
    def test_circuits_and_samples_per_circuit_need_to_have_the_same_length(
        self, n_samples_per_circuit
    ):
        circuits = [Circuit([H(0), CNOT(0, 1)]), Circuit([X(2)])]

        with pytest.raises(ValueError):
            split_into_batches(circuits, n_samples_per_circuit, 100)

    @pytest.mark.parametrize("max_circuits_per_batch", [-10, 0])
    def test_max_circuits_per_batch_has_to_be_positive(self, max_circuits_per_batch):
        circuits = [Circuit([X(0), CNOT(1, 2)]), Circuit([X(1), H(1)])]

        with pytest.raises(ValueError):
            split_into_batches(circuits, [100, 100], max_circuits_per_batch)


class TestBatching:
    @pytest.mark.parametrize("max_batch_size", [100, 1000])
    def test_one_item_is_returned_if_all_circuits_fit_into_single_batch(
        self, max_batch_size
    ):
        circuits = (Circuit([CNOT(0, 1)]), Circuit([X(2)]))

        batches = list(split_into_batches(circuits, [100, 50], max_batch_size))

        assert len(batches) == 1
        assert batches[0] == (circuits, 100)

    def test_batch_size_equals_to_max_of_requested_sample_sizes(self):
        circuits = (Circuit([CNOT(0, 1)]),) * 10
        samples_per_circuit = [5, 10, 21, 37, 12, 1, 10, 10, 10, 200]

        batches = list(
            split_into_batches(circuits, samples_per_circuit, max_batch_size=3)
        )

        assert batches == [
            (circuits[0:3], 21),
            (circuits[3:6], 37),
            (circuits[6:9], 10),
            (circuits[9:], 200),
        ]


class TestExpandingSampleSizes:
    def test_same_seqs_are_returned_if_all_sample_sizes_le_max_sample_size(self):
        circuits = [
            Circuit([X(0), CNOT(0, 1)]),
            Circuit([X(0), H(0)]),
            Circuit([H(0), H(1)]),
        ]
        n_samples_per_circuit = [20, 30, 40]

        new_circuits, new_n_samples, multiplicities = expand_sample_sizes(
            circuits, n_samples_per_circuit, 100
        )

        assert multiplicities == [1, 1, 1]
        assert new_n_samples == n_samples_per_circuit
        assert new_circuits == circuits

    def test_n_samples_and_multiplicities_are_correctly_computed(self):
        circuits = [Circuit([X(i)]) for i in range(6)]
        n_samples_per_circuit = [10, 20, 30, 20, 45, 40]
        max_sample_size = 20
        expected_new_circuits = [circuits[i] for i in (0, 1, 2, 2, 3, 4, 4, 4, 5, 5)]
        expected_new_n_samples = [10, 20, 20, 10, 20, 20, 20, 5, 20, 20]
        expected_multiplicities = [1, 1, 2, 1, 3, 2]

        new_circuits, new_n_samples, multiplicities = expand_sample_sizes(
            circuits, n_samples_per_circuit, max_sample_size
        )

        assert new_circuits == expected_new_circuits
        assert new_n_samples == expected_new_n_samples
        assert expected_multiplicities == multiplicities


class TestCombiningMeasurements:
    def test_raises_error_when_multiplicities_dont_match_measurements(self):
        multiplicities = [1, 2, 3, 2, 2]
        # Clearly a mismatch, we should have 10 Measurements object
        measurements = [{"00": 20, "11": 10} for _ in range(5)]

        with pytest.raises(ValueError):
            combine_measurement_counts(measurements, multiplicities)

    @pytest.mark.parametrize(
        "all_counts, multiplicities, combined_counts",
        [
            ([{"00": 10}], [1], [{"00": 10}]),
            (
                [
                    # Group 0
                    {"00": 5, "11": 3},
                    {"00": 4},
                    {"11": 2},
                    # Group 1
                    {"01": 3},
                    {"10": 4},
                    # Group 2
                    {"000": 20},
                    # Group 4
                    {"001": 3, "011": 2},
                    {"011": 2},
                    {"111": 10},
                ],
                [3, 2, 1, 3],
                [
                    {"00": 9, "11": 5},
                    {"01": 3, "10": 4},
                    {"000": 20},
                    {"001": 3, "011": 4, "111": 10},
                ],
            ),
        ],
    )
    def test_counts_of_combined_measurements_are_correct(
        self, all_counts, multiplicities, combined_counts
    ):
        assert all(
            actual == expected
            for actual, expected in zip(
                combine_measurement_counts(all_counts, multiplicities),
                combined_counts,
            )
        )


class TestCombiningBitstrings:
    def test_raises_error_when_multiplicities_dont_match_measurements(self):
        multiplicities = [1, 2, 3, 2, 2]
        # Clearly a mismatch, we should have 10 bitstring sequences
        measurements = [["00", "11"] for _ in range(5)]

        with pytest.raises(ValueError):
            combine_measurement_counts(measurements, multiplicities)

    @pytest.mark.parametrize(
        "all_bitstrings, multiplicities, combined_bitstrings",
        [
            ([["00", "11"]], [1], [["00", "11"]]),
            (
                [
                    ["00", "11"],
                    ["01", "00"],
                    ["00", "00"],
                    ["000"],
                    ["111", "001"],
                    ["00000"],
                    ["0", "0", "0"],
                    ["1", "0"],
                    ["0", "0"],
                ],
                [3, 2, 1, 3],
                [
                    ["00", "11", "01", "00", "00", "00"],
                    ["000", "111", "001"],
                    ["00000"],
                    ["0", "0", "0", "1", "0", "0", "0"],
                ],
            ),
        ],
    )
    def test_counts_of_combined_bitstrings_are_correct(
        self, all_bitstrings, multiplicities, combined_bitstrings
    ):
        assert all(
            actual == expected
            for actual, expected in zip(
                combine_bitstrings(all_bitstrings, multiplicities),
                combined_bitstrings,
            )
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Iterable, Sequence

from ..circuits._builtin_gates import RY, RZ
from ..circuits._circuit import Circuit
from ..circuits._gates import ControlledGate, GateOperation
from ..decompositions._decomposition import DecompositionRule, decompose_operations


class U3GateToRotation(DecompositionRule[GateOperation]):
    """Decomposition of Orquestra's U3 gate.

    Note that this gets rid of global phase.
    """

    def predicate(self, operation: GateOperation) -> bool:
        # Only decompose U3 and its controlled version
        return (
            operation.gate.name == "U3"
            or isinstance(operation.gate, ControlledGate)
            and operation.gate.wrapped_gate.name == "U3"
        )

    def production(self, operation: GateOperation) -> Iterable[GateOperation]:
        theta, phi, lambda_ = operation.params

        gate_decomposition = [RZ(phi), RY(theta), RZ(lambda_)]

        def preprocess_gate(gate):
            return (
                gate.controlled(operation.gate.num_control_qubits)
                if operation.gate.name == "Control"
                else gate
            )

        gate_operation_decomposition = [
            preprocess_gate(gate)(*operation.qubit_indices)
            for gate in gate_decomposition
        ]

        return reversed(gate_operation_decomposition)


def decompose_orquestra_circuit(
    circuit: Circuit, decomposition_rules: Sequence[DecompositionRule[GateOperation]]
):
    return Circuit(decompose_operations(circuit.operations, decomposition_rules))



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from functools import partial
from itertools import chain

import numpy as np

from ..circuits import CNOT, Circuit, H, X, builtin_gate_by_name
from ..operators import PauliTerm
from ..testing.test_cases_for_backend_tests import (
    one_qubit_non_parametric_gates_amplitudes_test_set,
    one_qubit_parametric_gates_amplitudes_test_set,
    two_qubit_non_parametric_gates_amplitudes_test_set,
    two_qubit_parametric_gates_amplitudes_test_set,
)
from .wavefunction_simulator import WavefunctionSimulator

_EXAMPLE_CIRCUITS = [
    Circuit([H(0)]),
    Circuit([H(0), CNOT(0, 1), CNOT(1, 2)]),
]

_CORRESPONDING_WAVEFUNCTIONS = [
    np.array([1, 1]) / np.sqrt(2),
    np.array([1, 0, 0, 0, 0, 0, 0, 1]) / np.sqrt(2),
]


_CORRESPONDING_OPERATORS = [PauliTerm("X0"), PauliTerm("X0 * Z1 * Y2")]


def _verify_wavefunction_returned_by_simulator_has_correct_length(
    simulator: WavefunctionSimulator,
):
    return all(
        len(simulator.get_wavefunction(circuit)) == 2**circuit.n_qubits
        for circuit in _EXAMPLE_CIRCUITS
    )


def _verify_wavefunction_returned_by_simulator_has_correct_coefficients(atol):
    def _contract(simulator):
        return all(
            np.allclose(simulator.get_wavefunction(circuit), wavefunction, atol=atol)
            for circuit, wavefunction in zip(
                _EXAMPLE_CIRCUITS, _CORRESPONDING_WAVEFUNCTIONS
            )
        )

    return _contract


def _verify_simulator_takes_into_account_initial_state_when_computing_wavefunction(
    atol,
):
    def _contract(simulator):
        circuit = Circuit([X(0)])
        return all(
            [
                np.allclose(
                    simulator.get_wavefunction(circuit, initial_state=np.array([0, 1])),
                    np.array([1, 0]),
                    atol=atol,
                ),
                np.allclose(
                    simulator.get_wavefunction(circuit, initial_state=np.array([1, 0])),
                    np.array([0, 1]),
                    atol=atol,
                ),
            ]
        )

    return _contract


def _verify_computing_wavefunction_increases_number_of_jobs_and_circuits_executed(
    simulator,
):
    n_jobs = [simulator.n_jobs_executed]
    n_circuits = [simulator.n_circuits_executed]

    for circuit in _EXAMPLE_CIRCUITS:
        simulator.get_wavefunction(circuit)
        n_jobs.append(simulator.n_jobs_executed)
        n_circuits.append(simulator.n_circuits_executed)

    return all(n_jobs[i] < n_jobs[i + 1] for i in range(len(n_jobs) - 1)) and all(
        n_circuits[i] < n_circuits[i + 1] for i in range(len(n_jobs) - 1)
    )


def _verify_computing_expectation_values_increases_number_of_jobs_and_circuits_executed(
    simulator,
):
    n_jobs = [simulator.n_jobs_executed]
    n_circuits = [simulator.n_circuits_executed]

    for circuit, operator in zip(_EXAMPLE_CIRCUITS, _CORRESPONDING_OPERATORS):
        simulator.get_exact_expectation_values(circuit, operator)
        n_jobs.append(simulator.n_jobs_executed)
        n_circuits.append(simulator.n_circuits_executed)

    return all(n_jobs[i] < n_jobs[i + 1] for i in range(len(n_jobs) - 1)) and all(
        n_circuits[i] < n_circuits[i + 1] for i in range(len(n_jobs) - 1)
    )


def _verify_gate_compatibility(simulator, circuit, target_amplitudes, atol):
    wavefunction = simulator.get_wavefunction(circuit)

    return np.allclose(wavefunction.amplitudes, target_amplitudes, atol=atol)


def _one_qubit_nonparametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gate)(0),
                    builtin_gate_by_name(tested_gate)(0),
                ]
            ),
            target_amplitudes,
        )
        for initial_gate, tested_gate, target_amplitudes in one_qubit_non_parametric_gates_amplitudes_test_set  # noqa: E501
        if tested_gate not in gates_to_exclude
    ]


def _two_qubit_nonparametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gates[0])(0),
                    builtin_gate_by_name(initial_gates[1])(1),
                    builtin_gate_by_name(tested_gate)(0, 1),
                ]
            ),
            target_amplitudes,
        )
        for initial_gates, tested_gate, target_amplitudes in two_qubit_non_parametric_gates_amplitudes_test_set  # noqa: E501
        if tested_gate not in gates_to_exclude
    ]


def _one_qubit_parametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gate)(0),
                    builtin_gate_by_name(tested_gate)(*params)(0),
                ]
            ),
            target_amplitudes,
        )
        for initial_gate, tested_gate, params, target_amplitudes in one_qubit_parametric_gates_amplitudes_test_set  # noqa: E501
        if tested_gate not in gates_to_exclude
    ]


def _two_qubit_parametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gates[0])(0),
                    builtin_gate_by_name(initial_gates[1])(1),
                    builtin_gate_by_name(tested_gate)(*params)(0, 1),
                ]
            ),
            target_amplitudes,
        )
        for initial_gates, tested_gate, params, target_amplitudes in two_qubit_parametric_gates_amplitudes_test_set  # noqa: E501
        if tested_gate not in gates_to_exclude
    ]


def simulator_contracts_for_tolerance(atol=1e-7):
    return [
        _verify_wavefunction_returned_by_simulator_has_correct_length,
        _verify_wavefunction_returned_by_simulator_has_correct_coefficients(atol),
        _verify_computing_wavefunction_increases_number_of_jobs_and_circuits_executed,
        _verify_computing_expectation_values_increases_number_of_jobs_and_circuits_executed,  # noqa: E501
    ]


def simulator_contracts_with_nontrivial_initial_state(atol=1e-7):
    return [
        _verify_simulator_takes_into_account_initial_state_when_computing_wavefunction(
            atol
        ),
    ]


def simulator_gate_compatibility_contracts(atol=1e-7, gates_to_exclude=None):
    gates_to_exclude = [] if gates_to_exclude is None else gates_to_exclude

    return [
        partial(
            _verify_gate_compatibility,
            circuit=circuit,
            target_amplitudes=target_amplitudes,
            atol=atol,
        )
        for circuit, target_amplitudes in chain(
            _one_qubit_nonparametric_gate_test_cases(gates_to_exclude),
            _two_qubit_nonparametric_gate_test_cases(gates_to_exclude),
            _one_qubit_parametric_gate_test_cases(gates_to_exclude),
        )
    ]



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import warnings

import numpy as np
import pytest
import sympy

from orquestra.quantum.circuits import MultiPhaseOperation


class TestMultiPhaseOperation:
    @pytest.mark.parametrize(
        "wavefunction",
        [
            np.array([1, 0, 0, 0]),
            np.array([1, 1, 0, 0]) / np.sqrt(2),
            np.array([1, 0, 0, 1]) / np.sqrt(2),
            np.array([1j, 1j, 1j, 1j]) * 0.5,
            np.array([0, 1]),
            np.array([1, 0, 1, 0, 0, 1, 1, 0]) / 2,
        ],
    )
    def test_with_all_parameters_set_to_0_behaves_like_identity(self, wavefunction):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", np.ComplexWarning)
            params = tuple(np.zeros_like(wavefunction))
            operation = MultiPhaseOperation(params)

            np.testing.assert_array_equal(operation.apply(wavefunction), wavefunction)

    @pytest.mark.parametrize(
        "wavefunction,params",
        [
            (np.array([1, 0, 0, 0]), ()),
            (np.array([1, 1, 0, 0]) / np.sqrt(2), (0, 1, 2)),
            (np.array([1, 0, 0, 1]) / np.sqrt(2), (0, 1, 2, 3, 4)),
        ],
    )
    def test_cannot_be_applied_if_length_of_params_and_wavefunction_dont_match(
        self, wavefunction, params
    ):
        operation = MultiPhaseOperation(params)
        with pytest.raises(ValueError):
            operation.apply(wavefunction)

    @pytest.mark.parametrize(
        "wavefunction,params",
        [
            (np.array([1, 1, 1, 1]) / 2, (3j, 1, 2, 0)),
            (np.array([1, 1, 1, 1]) / 2, (0, 1 + 1j, 2, 3j)),
        ],
    )
    def test_cannot_be_applied_if_params_are_not_real(self, wavefunction, params):
        with pytest.raises(ValueError):
            MultiPhaseOperation(params)

    @pytest.mark.parametrize(
        "wavefunction,params,target_wavefunction",
        [
            [
                np.array([1, 1, 1, 1]) / 2,
                np.array([np.pi, -np.pi / 2, np.pi / 3, -np.pi / 4]),
                np.array(
                    [
                        np.exp(np.pi * 1j),
                        np.exp(-np.pi / 2 * 1j),
                        np.exp(np.pi / 3 * 1j),
                        np.exp(-np.pi / 4 * 1j),
                    ]
                )
                / 2,
            ],
            [
                np.array([1, 0, 0, 1]) / np.sqrt(2),
                np.array([np.pi, np.pi / 2, np.pi / 3, np.pi / 4]),
                np.array(
                    [
                        np.exp(np.pi * 1j),
                        0,
                        0,
                        np.exp(np.pi / 4 * 1j),
                    ]
                )
                / np.sqrt(2),
            ],
            [
                np.array([1, 1, 1, 1]) / 2,
                np.array([np.pi, 0, 0, 0]),
                np.array(
                    [
                        np.exp(np.pi * 1j),
                        1,
                        1,
                        1,
                    ]
                )
                / 2,
            ],
            [
                np.array([1, 1, 1, 1]) / 2,
                np.array([sympy.pi, sympy.sympify(0), 0, 0]),
                np.array(
                    [
                        np.exp(np.pi * 1j),
                        1,
                        1,
                        1,
                    ]
                )
                / 2,
            ],
        ],
    )
    def test_correctly_transforms_wavefunction(
        self, wavefunction, params, target_wavefunction
    ):
        operation = MultiPhaseOperation(params)
        output = operation.apply(wavefunction)
        np.testing.assert_allclose(output, target_wavefunction)

    @pytest.mark.parametrize(
        "wavefunction",
        [np.array([1, 1, 1, 1]) / 2, np.array([1, 0, 0, 1]) / np.sqrt(2)],
    )
    @pytest.mark.parametrize(
        "params", [tuple(sympy.symbols("alpha_0:4")), (sympy.Symbol("theta"), 1, 1, 1)]
    )
    def test_apply_does_not_work_with_symbolic_params(self, wavefunction, params):
        operation = MultiPhaseOperation(params)
        with pytest.raises(RuntimeError):
            operation.apply(wavefunction)

    @pytest.mark.parametrize(
        "params, expected_free_symbols",
        [
            ((0, 1, 1, -2), []),
            (
                (sympy.Symbol("a"), sympy.Symbol("b")),
                [sympy.Symbol("a"), sympy.Symbol("b")],
            ),
            (
                (
                    1,
                    sympy.Symbol("a") + sympy.Symbol("d"),
                    -1,
                    sympy.Symbol("c") - sympy.Symbol("b"),
                ),
                list(sympy.symbols("a, b, c, d")),
            ),
        ],
    )
    def test_free_symbols_in_parameters_are_correctly_reported(
        self, params, expected_free_symbols
    ):
        operation = MultiPhaseOperation(params)
        assert operation.free_symbols == expected_free_symbols



#! /usr/bin/env python3
################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################

# This script ia intended to be run from within a Github Action.
# Reads current project version, bumps "minor", and sets the "next_version" output
# variable using Github's special stdout syntax.

import sys
import re
import typing as t
from importlib.metadata import version


class Semver(t.NamedTuple):
    major: int
    minor: int
    patch: t.Optional[int]
    pre: t.Optional[str]

    @property
    def version_str(self):
        version = f"{self.major}.{self.minor}"
        if self.patch is not None:
            version += f".{self.patch}"
        if self.pre:
            version += f".{self.pre}"

        return version

    @property
    def bump_minor(self):
        return Semver(
            major=self.major,
            minor=self.minor + 1,
            patch=0,
            pre=None,
        )


SEMVER_REGEX = (
    r"(?P<major>[0-9]+)\.(?P<minor>[0-9]+)(\.(?P<patch>[0-9]+))?([-\.](?P<pre>.+))?"
)


def parse_version_str(version: str) -> Semver:
    # TODO: make it more resilient to versions with missing components if we need it.
    # We should be fine for some time, because setuptools_scm return all components.
    match = re.match(SEMVER_REGEX, version)
    if match is None:
        raise ValueError(f"Can't parse version string '{version}'")

    groups = match.groupdict()

    # re sets up `None` as the value for a missing group
    if groups["patch"] is not None:
        patch = int(groups["patch"])
    else:
        patch = None

    return Semver(
        major=int(groups["major"]),
        minor=int(groups["minor"]),
        patch=patch,
        pre=groups.get("pre"),
    )


def _set_github_output(name: str, value):
    # Special Github syntax for setting outputs from steps in Github Actions. See:
    # https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-an-output-parameter
    print(f"::set-output name={name}::{value}")


def main(package_name):
    """Run the actual script logic.

    Args:
        version_verride: should contain Github Action input. Empty strings are treated
            as "nil" values. This is because that's how bash passes nils.
    """

    current_version = version(package_name)
    print(f"Read current version as: {current_version}")

    next_version = parse_version_str(current_version).bump_minor.version_str

    print(f"Next version: {next_version}")
    _set_github_output("next_version", next_version)


if __name__ == "__main__":
    package_name = sys.argv[1]
    main(package_name=package_name)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from ._estimation import (
    calculate_exact_expectation_values,
    estimate_expectation_values_by_averaging,
    evaluate_estimation_circuits,
    evaluate_non_measured_estimation_tasks,
    split_estimation_tasks_to_measure,
)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Public API for the orquestra circuits.

Allows:

- defining quantum circuits with gates applied to qubits
- set of built-in gates (see imports in this `__init__`)
- circuit (de)serialization to/from JSON-compatible dicts

Examples:

    Defining a circuit with a NOT gate on qubit 0 & Hadamard gate on qubit 1:

    .. code:: python

        circuit = Circuit()
        circuit += X(0)
        circuit += H(1)

    Adding 2-qubit gates:

    .. code:: python

        circuit += CNOT(0, 4)

    Adding parametrized gates:

    .. code:: python

        circuit += RX(np.pi / 2)(0)
        circuit += RX(sympy.sympify("theta / 2"))(1)

    Adding 2-qubit parametrized gates:

    .. code:: python

        circuit += CPHASE(0.314)(3, 2)

    Adding a built-in gate by its name:

    .. code:: python

        circuit += builtin_gate_by_name("X")(0)
        circuit += builtin_gate_by_name("RX")(np.pi * 1.5)(1)

    Binding parameters:

    .. code:: python

        circuit = circuit.bind({sympy.Symbol("theta"): -np.pi / 5})

    Iterating over circuit contents:

    .. code:: python

        for gate_op in circuit.operations:
            name = gate_op.gate.name
            params = gate_op.gate.params
            qubits = gate_op.qubit_indices
            print(f"{name} with params {params} applied to {qubits}")

    Making a different circuit (e.g. shifting gates by 1 qubit):

    .. code:: python

        new_circuit = Circuit(
            operations=[
                gate_op.gate(*[qubit + 1 for qubit in gate_op.qubits])
                for gate_op in circuit.operations
            ],
            n_qubits=circuit.n_qubits
        )

    (De)serialization:

    .. code:: python

        to_dict(circuit)
        circuit5 = circuit_from_dict(dict5)

    Inverting a circuit::

    .. code:: python

        circuit_inverted = circuit.invert()


Defining new gates
------------------

To use a gate that isn't already covered by built-in ones you can define a custom gate
or extend the set of the built-in ones and file a PR to orquestra-quantum.

Using custom gates:

.. code:: python

    custom_a = circuits.CustomGateDefinition(
        gate_name="custom_a",  # names need to be unique
        matrix=sympy.Matrix(
            [
                [-1, 0],
                [0, 1],
            ]
        ),
        params_ordering=(),
    )

    custom_b = circuits.CustomGateDefinition(
        gate_name="custom_b",
        matrix=sympy.Matrix(
            [
                [0, sympy.Symbol("theta") * 2],
                [sympy.Symbol("gamma") + 3, 1],
            ]
        ),
        params_ordering=(sympy.Symbol("gamma"), sympy.Symbol("theta")),
    )

    circuit = Circuit()
    circuit += custom_a()(0)
    circuit += custom_b(np.pi, np.pi / 2)(0)


Extending built-in gates requires:

- Adding its definition to `orquestra.quantum.circuits._builtin_gates`. Refer to other
    1- or multi-qubit, parametric/nonparametric gates there to see how it's been done
    for other gates.

- Adding its matrix to `orquestra.quantum.circuits._matrices`.

- Adding tests for conversion to other frameworks in:
    - `qeqiskit.conversions.circuit_conversions_test`
    - `qecirq.conversions.circuit_conversions_test`
    - `qeforest.conversions.circuit_conversions_test`

- Implement conversions. Some might work out of the box, e.g. if there's a gate with the
    same name defined in PyQuil our converters will use it by default without need for
    explicit mappings.
"""

from ._builtin_gates import (
    CNOT,
    CPHASE,
    CZ,
    ISWAP,
    MS,
    PHASE,
    RH,
    RX,
    RY,
    RZ,
    SWAP,
    U3,
    XX,
    XY,
    YY,
    ZZ,
    Delay,
    GatePrototype,
    GPi,
    GPi2,
    H,
    I,
    S,
    T,
    X,
    Y,
    Z,
    builtin_gate_by_name,
)
from ._circuit import Circuit, split_circuit
from ._gates import (
    ControlledGate,
    CustomGateDefinition,
    Dagger,
    Exponential,
    Gate,
    GateOperation,
    MatrixFactoryGate,
)
from ._generators import (
    add_ancilla_register,
    apply_gate_to_qubits,
    create_layer_of_gates,
)
from ._itertools import (
    combine_bitstrings,
    combine_measurement_counts,
    expand_sample_sizes,
    split_into_batches,
)
from ._operations import Operation
from ._serde import (
    circuit_from_dict,
    circuitset_from_dict,
    load_circuit,
    load_circuitset,
    save_circuit,
    save_circuitset,
    to_dict,
)
from ._testing import create_random_circuit
from ._wavefunction_operations import MultiPhaseOperation
from .symbolic import natural_key, natural_key_revlex



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import unittest

from orquestra.quantum.distributions import MeasurementOutcomeDistribution
from orquestra.quantum.distributions.BAS_dataset import (
    bars_and_stripes_zigzag,
    get_bars_and_stripes_target_distribution,
    get_num_bars_and_stripes_patterns,
)


class TestTarget(unittest.TestCase):
    def setUp(self):
        self.methods = ["zigzag"]
        self.grids = [[1, 1], [1, 2], [2, 1], [2, 2], [3, 2], [2, 3], [3, 3]]
        self.number_of_patterns = [2, 4, 4, 6, 10, 10, 14]
        self.fractions = [1.0, 0.75, 0.5, 0.25, 0.1]

    def test_get_num_bars_and_stripes_patterns(self):
        for i, grid in enumerate(self.grids):
            # When
            given_num_patterns = get_num_bars_and_stripes_patterns(grid[0], grid[1])

            # Then
            self.assertEqual(given_num_patterns, self.number_of_patterns[i])

    def test_get_bars_and_stripes_target_distribution_all_methods(self):
        for method in self.methods:
            for i, grid in enumerate(self.grids):
                for fraction in self.fractions:
                    # When
                    bars_and_stripes_distribution = (
                        get_bars_and_stripes_target_distribution(
                            grid[0], grid[1], fraction=fraction, method=method
                        )
                    )

                    # Then
                    self.assertIsInstance(
                        bars_and_stripes_distribution, MeasurementOutcomeDistribution
                    )
                    expected_num_patters = max(
                        int(self.number_of_patterns[i] * fraction), 1
                    )
                    self.assertEqual(
                        len(bars_and_stripes_distribution.distribution_dict.keys()),
                        expected_num_patters,
                    )

    def test_get_bars_and_stripes_target_distribution_unsupported_method(self):
        # Given
        method = "METHOD TYPE NOT SUPPORTED"
        nrows = 2
        ncols = 2

        # When/Then
        self.assertRaises(
            RuntimeError,
            lambda: get_bars_and_stripes_target_distribution(
                nrows, ncols, fraction=1.0, method=method
            ),
        )

    def test_bars_and_stripes_zigzag_2_by_2(self):
        # Given
        ncols = 2
        nrows = 2
        expected_patterns = ["0000", "0011", "1100", "1010", "0101", "1111"]

        # When
        pattern_list = bars_and_stripes_zigzag(nrows, ncols)

        # Then
        self.assertEqual(len(expected_patterns), len(pattern_list))
        for pattern in pattern_list:
            bitstring = ""
            for qubit in pattern:
                bitstring += str(qubit)
            self.assertIn(bitstring, expected_patterns)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import random
from typing import List, Optional

import numpy as np

from ..utils import ValueEstimate, bin2dec, dec2bin
from ..wavefunction import Wavefunction
from ._openfermion_utils.sparse_tools import expectation, get_sparse_operator
from ._pauli_operators import PauliRepresentation, PauliSum, PauliTerm


def get_pauliop_from_matrix(operator: List[List]) -> PauliSum:
    r"""Expands a 2^n by 2^n matrix into n-qubit Pauli basis. The runtime of
    this function is O(2^2n).

    Args:
        operator: a list of lists (rows) representing a 2^n by 2^n
            matrix.

    Returns:
        A PauliSum instance corresponding to the expansion of
        the input operator as a sum of Pauli strings:

        O = 2^-n \sum_P tr(O*P) P
    """

    nrows = len(operator)
    ncols = len(operator[0])

    # Check if the input operator is square
    if nrows != ncols:
        raise Exception("The input operator is not square")

    # Check if the dimensions are powers of 2
    if not (((nrows & (nrows - 1)) == 0) and nrows > 0):
        raise Exception("The number of rows is not a power of 2")
    if not (((ncols & (ncols - 1)) == 0) and ncols > 0):
        raise Exception("The number of cols is not a power of 2")

    n = int(np.log2(nrows))  # number of qubits

    def decode(bit_string):  # Helper function for converting any 2n-bit
        # string to a label vector representing a Pauli
        # string of length n

        if len(bit_string) != 2 * n:
            raise Exception("LH_expand:decode: input bit string length not 2n")

        output_label = list(np.zeros(n))
        for i in range(0, n):
            output_label[i] = bin2dec(bit_string[2 * i : 2 * i + 2])

        return output_label

    def trace_product(label_vec):  # Helper function for computing tr(OP)
        # where O is the input operator and P is a
        # Pauli string operator

        def f(j):  # Function which computes the index of the nonzero
            # element in P for a given column j

            j_str = dec2bin(j, n)
            for index in range(0, n):
                if label_vec[index] in [1, 2]:  # flip if X or Y
                    j_str[index] = int(not j_str[index])
            return bin2dec(j_str)

        def nz(j):  # Function which computes the value of the nonzero
            # element in P on the column j

            val_nz = 1.0
            j_str = dec2bin(j, n)
            for index in range(0, n):
                if label_vec[index] == 2:
                    if j_str[index] == 0:
                        val_nz = val_nz * (1j)
                    if j_str[index] == 1:
                        val_nz = val_nz * (-1j)
                if label_vec[index] == 3:
                    if j_str[index] == 1:
                        val_nz = val_nz * (-1)
            return val_nz

        # Compute the trace
        tr = 0.0
        for j in range(0, 2**n):  # loop over the columns
            tr = tr + operator[j][f(j)] * nz(j)

        return tr / 2**n

    # Expand the operator in Pauli basis
    coeffs = list(np.zeros(4**n))
    labels = list(np.zeros(4**n))
    for i in range(0, 4**n):  # loop over all 2n-bit strings
        current_string = dec2bin(i, 2 * n)  # see util.py
        current_label = decode(current_string)
        coeffs[i] = trace_product(current_label)
        labels[i] = current_label

    return get_pauliop_from_coeffs_and_labels(coeffs, labels)


def get_pauliop_from_coeffs_and_labels(
    coeffs: List[float], labels: List[List[int]]
) -> PauliSum:
    """Generates a PauliSum based on a coefficient vector and
    a label matrix.

    Args:
        coeffs: a list of floats representing the coefficients
            for the terms in the Hamiltonian
        labels: a list of lists (a matrix) where each list
            is a vector of integers representing the Pauli
            string. See pauliutil.py for details.

    Example:

        The Hamiltonian H = 0.1 X1 X2 - 0.4 Y1 Y2 Z3 Z4 can be
        initiated by calling

        H = get_pauliop_from_coeffs_and_labels([0.1, -0.4],  # coefficients
            [[1 1 0 0],  # label matrix
            [2 2 3 3]])
    """

    output = PauliSum()
    for i in range(0, len(labels)):
        string_term = ""
        for ind, elem in enumerate(labels[i]):
            pauli_symbol = ""
            if elem == 1:
                pauli_symbol = "*X" + str(ind)
            elif elem == 2:
                pauli_symbol = "*Y" + str(ind)
            elif elem == 3:
                pauli_symbol = "*Z" + str(ind)
            string_term += pauli_symbol

        output += PauliTerm(f"{coeffs[i]}{string_term}")

    return output


def generate_random_pauliop(
    nqubits: int,
    nterms: int,
    nlocality: int,
    max_coeff: float,
    fixed_coeff: bool = False,
) -> PauliSum:
    """Generates a Hamiltonian with term coefficients uniformly distributed
    in [-max_coeff, max_coeff].

    Args:
        nqubits   - number of qubits
        nterms    - number of terms in the Hamiltonian
        nlocality - locality of the Hamiltonian
        max_coeff - bound for generating the term coefficients
        fixed_coeff (bool) - If true, all the terms are assign the
            max_coeff as coefficient.

    Returns:
        A PauliSum with the appropriate coefficient vector
        and label matrix.
    """
    # generate random coefficient vector
    if fixed_coeff:
        coeffs = [max_coeff] * nterms
    else:
        coeffs = list(np.zeros(nterms))
        for j in range(0, nterms):
            coeffs[j] = random.uniform(-max_coeff, max_coeff)

    # generate random label vector
    labels = list(np.zeros(nterms, dtype=int))
    label_set = set()
    j = 0
    while j < nterms:
        inds_nontrivial = sorted(random.sample(range(0, nqubits), nlocality))
        label = list(np.zeros(nqubits, dtype=int))
        for ind in inds_nontrivial:
            label[ind] = random.randint(1, 3)
        if str(label) not in label_set:
            labels[j] = label
            j += 1
            label_set.add(str(label))
    return get_pauliop_from_coeffs_and_labels(coeffs, labels)


def evaluate_operator(
    operator: PauliRepresentation, expectation_values
) -> ValueEstimate:
    """Evaluate the expectation value of a qubit operator using expectation values for
    the terms.

    Args:
        operator: the operator
        expectation_values: the expectation values

    Returns:
        value_estimate: stores the value of the expectation and its precision
    """

    # Sum the contributions from all terms
    total = 0

    # Add all non-trivial terms
    for i, term in enumerate(operator.terms):
        total += np.real(term.coefficient * expectation_values.values[i])

    value_estimate = ValueEstimate(total)
    return value_estimate


def evaluate_operator_list(
    operator_list: List[PauliRepresentation],
    expectation_values,
) -> ValueEstimate:
    """Evaluate the expectation value of an operator list using expectation values for
    the terms. The expectation values should be in the order given by the qubit operator
    list, and the value returned is the sum of all terms in the qubit operator list.

    Args:
        operator_list: the operator list
        expectation_values: the expectation values

    Returns:
        value_estimate: stores the value of the expectation and its precision
    """

    # Sum the contributions from all terms
    total = 0

    # Add all non-trivial terms
    term_index = 0
    for operator in operator_list:
        for term in operator.terms:
            total += np.real(term.coefficient * expectation_values.values[term_index])
            term_index += 1

    value_estimate = ValueEstimate(total)
    return value_estimate


def reverse_qubit_order(
    qubit_operator: PauliRepresentation, n_qubits: Optional[int] = None
):
    """Reverse the order of qubit indices in a qubit operator.

    Args:
        qubit_operator: the operator to be reversed
        n_qubits (int): total number of qubits. Needs to be provided when
            the size of the system of interest is greater than the size of qubit
            operator (optional)

    Returns:
        reversed_op: the reversed operator
    """

    reversed_op = PauliSum()

    if n_qubits is None:
        n_qubits = qubit_operator.n_qubits
    if n_qubits < qubit_operator.n_qubits:
        raise ValueError("Invalid number of qubits specified.")

    for term in qubit_operator.terms:
        new_term = {}
        for qubit_num, operator_str in term.operations:
            new_qubit_num = n_qubits - 1 - qubit_num
            new_term[new_qubit_num] = operator_str
        reversed_op += PauliTerm(new_term, term.coefficient)
    return reversed_op


def get_expectation_value(
    qubit_op: PauliRepresentation,
    wavefunction: Wavefunction,
    reverse_operator: bool = False,
) -> complex:
    """Get the expectation value of a qubit operator with respect to a wavefunction.

    Args:
        qubit_op: the operator
        wavefunction: the wavefunction
        reverse_operator: whether to reverse order of qubit operator
            before computing expectation value. This should be True if the convention
            of the basis states used for the wavefunction is the opposite of the one in
            the qubit operator. This is the case when the wavefunction uses
            Rigetti convention (https://arxiv.org/abs/1711.02086) of ordering qubits.
    Returns:
        the expectation value
    """
    n_qubits = wavefunction.amplitudes.shape[0].bit_length() - 1

    # Convert the qubit operator to a sparse matrix. Note that the qubit indices
    # must be reversed because OpenFermion and our Wavefunction use
    # different conventions for how to order the computational basis states!
    if reverse_operator:
        qubit_op = reverse_qubit_order(qubit_op, n_qubits=n_qubits)
    sparse_op = get_sparse_operator(qubit_op, n_qubits=n_qubits)

    # Computer the expectation value
    exp_val = expectation(sparse_op, wavefunction.amplitudes)
    return exp_val



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from operator import mul

import sympy
from sympy.physics.quantum import TensorProduct

# This script is used to generate test cases for tests in
# `quantum/interfaces/backend_test.py`. In order to use just run this python file with
# python3 `generate_cases_for_backend_tests.py` and then copy-paste the console output
# to the appropriate place in `test_cases_for_backend_tests.py`

# If someone would like to add new testcases to quantum/api/backend_test.py, e.g.
# testing new gate or different initial state, they could use this script to do that.
# (or they can do that manually).

I = sympy.Matrix([[1, 0], [0, 1]])  # noqa: E741
H = sympy.Matrix(
    [[1 / sympy.sqrt(2), 1 / sympy.sqrt(2)], [1 / sympy.sqrt(2), -1 / sympy.sqrt(2)]]
)
X = sympy.Matrix([[0, 1], [1, 0]])
Y = sympy.Matrix([[0, -sympy.I], [sympy.I, 0]])
Z = sympy.Matrix([[1, 0], [0, -1]])
S = sympy.Matrix([[1, 0], [0, sympy.I]])
T = sympy.Matrix([[1, 0], [0, (1 + sympy.I) / sympy.sqrt(2)]])

CNOT = sympy.Matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
CZ = sympy.Matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
SWAP = sympy.Matrix([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
ISWAP = sympy.Matrix([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]])

II = TensorProduct(I, I)
IH = TensorProduct(I, H)
HI = TensorProduct(H, I)

HH = TensorProduct(H, H)
XX = TensorProduct(X, X)
YY = TensorProduct(Y, Y)
ZZ = TensorProduct(Z, Z)
IX = TensorProduct(I, X)
ZI = TensorProduct(Z, I)

single_qubit_initial_states = [(I, "I"), (H, "H")]
single_qubit_operators = [I, X, Y, Z]

two_qubit_initial_states = [
    (II, ["I", "I"]),
    (IH, ["I", "H"]),
    (HI, ["H", "I"]),
    (HH, ["H", "H"]),
]
two_qubit_operators = [
    (II, "[]"),
    (XX, "[X0 X1]"),
    (YY, "[Y0 Y1]"),
    (ZZ, "[Z0 Z1]"),
    (IX, "[X1]"),
    (ZI, "[Z0]"),
]


def generate_cases_1_qubit_wavefunction(matrix, matrix_name, angles):
    for initial_matrix, initial_matrix_name in single_qubit_initial_states:
        for angle in angles:
            circuit = mul(matrix, initial_matrix)
            new_circuit = circuit.subs("theta", angle)
            amplitudes = new_circuit * sympy.Matrix([[1], [0]])
            gate_names_string = '["{}", "{}", '.format(initial_matrix_name, matrix_name)
            angle_string = "[{}], ".format(angle).replace("pi", "np.pi")
            amplitude_string = (
                str([amplitudes[0], amplitudes[1]])
                .replace("sqrt", "np.sqrt")
                .replace("pi", "np.pi")
                .replace("1.0*I", "1.0j")
                .replace("*I", "*1.0j")
                .replace("exp", "np.exp")
            )
            print(gate_names_string + angle_string + amplitude_string + "],")


def generate_cases_1_qubit_exp_vals(matrix, matrix_name):
    for initial_matrix, initial_matrix_name in single_qubit_initial_states:
        outputs = []
        for operator in single_qubit_operators:
            circuit_ket = mul(matrix, initial_matrix) * sympy.Matrix([[1], [0]])
            circuit_bra = sympy.conjugate(sympy.Transpose(circuit_ket))
            expectation_value = mul(circuit_bra, mul(operator, circuit_ket))[0]
            outputs.append(sympy.simplify(expectation_value))

        gate_names_string = '["{}", "{}", '.format(initial_matrix_name, matrix_name)
        exp_vals_string = (
            "[{}, {}, {}, {}]".format(*outputs)
            .replace("sqrt", "np.sqrt")
            .replace("pi", "np.pi")
            .replace("1.0*I", "1.0j")
            .replace("*I", "*1.0j")
        )
        print(gate_names_string + exp_vals_string + "],")


def generate_cases_1_qubit_exp_vals_with_angles(matrix, matrix_name, angles):
    for initial_matrix, initial_matrix_name in single_qubit_initial_states:
        for angle in angles:
            outputs = []
            for operator in single_qubit_operators:
                circuit_ket = mul(matrix, initial_matrix) * sympy.Matrix([[1], [0]])
                circuit_ket = circuit_ket.subs("theta", angle)
                circuit_bra = sympy.conjugate(sympy.Transpose(circuit_ket))
                expectation_value = mul(circuit_bra, mul(operator, circuit_ket))[0]
                outputs.append(sympy.simplify(expectation_value))

            gate_names_string = '["{}", "{}", '.format(initial_matrix_name, matrix_name)
            angle_string = "[{}], ".format(angle).replace("pi", "np.pi")
            exp_vals_string = (
                "[{}, {}, {}, {}]".format(*outputs)
                .replace("sqrt", "np.sqrt")
                .replace("pi", "np.pi")
                .replace("1.0*I", "1.0j")
                .replace("*I", "*1.0j")
            )
            print(gate_names_string + angle_string + exp_vals_string + "],")


def generate_cases_2_qubits_wavefunction(matrix, matrix_name, angles):
    for initial_matrix, initial_matrix_names in two_qubit_initial_states:
        for angle in angles:
            circuit = mul(matrix, initial_matrix)
            new_circuit = circuit.subs("theta", angle)
            amplitudes = new_circuit * sympy.Matrix([[1], [0], [0], [0]])
            gate_names_string = '[["{}", "{}"], "{}", '.format(
                initial_matrix_names[0], initial_matrix_names[1], matrix_name
            )
            angle_string = "[{}], ".format(angle).replace("pi", "np.pi")
            amplitude_string = (
                str([amplitudes[0], amplitudes[1], amplitudes[2], amplitudes[3]])
                .replace("sqrt", "np.sqrt")
                .replace("pi", "np.pi")
                .replace("1.0*I", "1.0j")
                .replace("*I", "*1.0j")
            )
            print(gate_names_string + angle_string + amplitude_string + "],")


def generate_cases_2_qubits_exp_vals(matrix, matrix_name):
    for initial_matrix, initial_matrix_names in two_qubit_initial_states:
        outputs = []
        operator_names = []
        for operator, operator_name in two_qubit_operators:
            operator_names.append('"{}"'.format(operator_name))
            circuit_ket = mul(matrix, initial_matrix) * sympy.Matrix(
                [[1], [0], [0], [0]]
            )
            circuit_bra = sympy.conjugate(sympy.Transpose(circuit_ket))
            expectation_value = mul(circuit_bra, mul(operator, circuit_ket))[0]

            outputs.append(sympy.simplify(expectation_value))
        operator_names_string = "[" + ", ".join(operator_names) + "], "
        gate_names_string = '[["{}", "{}"], "{}", '.format(
            initial_matrix_names[0], initial_matrix_names[1], matrix_name
        )
        exp_vals_string = "["
        for output in outputs:
            exp_vals_string += (
                "{},".format(output)
                .replace("sqrt", "np.sqrt")
                .replace("pi", "np.pi")
                .replace("1.0*I", "1.0j")
                .replace("*I", "*1.0j")
            )
        exp_vals_string += "]"

        print(gate_names_string + operator_names_string + exp_vals_string + "],")


def generate_cases_2_qubits_exp_vals_with_angles(matrix, matrix_name, angles):
    for initial_matrix, initial_matrix_names in two_qubit_initial_states:
        for angle in angles:
            outputs = []
            operator_names = []
            for operator, operator_name in two_qubit_operators:
                operator_names.append('"{}"'.format(operator_name))
                circuit_ket = mul(matrix, initial_matrix) * sympy.Matrix(
                    [[1], [0], [0], [0]]
                )
                circuit_ket = circuit_ket.subs("theta", angle)
                circuit_bra = sympy.conjugate(sympy.Transpose(circuit_ket))
                expectation_value = mul(circuit_bra, mul(operator, circuit_ket))[0]

                outputs.append(sympy.simplify(expectation_value))
            operator_names_string = "[" + ", ".join(operator_names) + "], "
            gate_names_string = '[["{}", "{}"], "{}", '.format(
                initial_matrix_names[0], initial_matrix_names[1], matrix_name
            )
            angle_string = "[{}], ".format(angle).replace("pi", "np.pi")
            exp_vals_string = "["
            for output in outputs:
                exp_vals_string += (
                    "{},".format(output)
                    .replace("sqrt", "np.sqrt")
                    .replace("pi", "np.pi")
                    .replace("1.0*I", "1.0j")
                    .replace("*I", "*1.0j")
                )
            exp_vals_string += "]"

            print(
                gate_names_string
                + angle_string
                + operator_names_string
                + exp_vals_string
                + "],"
            )


def main():
    theta = sympy.Symbol("theta")
    Rx = sympy.Matrix(
        [
            [sympy.cos(theta / 2), -1j * sympy.sin(theta / 2)],
            [-1j * sympy.sin(theta / 2), sympy.cos(theta / 2)],
        ]
    )
    Ry = sympy.Matrix(
        [
            [sympy.cos(theta / 2), -sympy.sin(theta / 2)],
            [sympy.sin(theta / 2), sympy.cos(theta / 2)],
        ]
    )
    Rz = sympy.Matrix(
        [
            [sympy.cos(theta / 2) - 1j * sympy.sin(theta / 2), 0],
            [0, sympy.cos(theta / 2) + 1j * sympy.sin(theta / 2)],
        ]
    )
    PHASE = sympy.Matrix([[1, 0], [0, sympy.cos(theta) + 1j * sympy.sin(theta)]])
    RH_phase_factor = sympy.exp(1j * theta / 2)
    RH = sympy.Matrix(
        [
            [
                RH_phase_factor
                * (sympy.cos(theta / 2) - 1j / sympy.sqrt(2) * sympy.sin(theta / 2)),
                RH_phase_factor * -1j / sympy.sqrt(2) * sympy.sin(theta / 2),
            ],
            [
                RH_phase_factor * -1j / sympy.sqrt(2) * sympy.sin(theta / 2),
                RH_phase_factor
                * (sympy.cos(theta / 2) + 1j / sympy.sqrt(2) * sympy.sin(theta / 2)),
            ],
        ]
    )

    CPHASE = sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, sympy.cos(theta) + 1j * sympy.sin(theta)],
        ]
    )
    XX = sympy.Matrix(
        [
            [sympy.cos(theta / 2), 0, 0, -1j * sympy.sin(theta / 2)],
            [0, sympy.cos(theta / 2), -1j * sympy.sin(theta / 2), 0],
            [0, -1j * sympy.sin(theta / 2), sympy.cos(theta / 2), 0],
            [-1j * sympy.sin(theta / 2), 0, 0, sympy.cos(theta / 2)],
        ]
    )
    YY = sympy.Matrix(
        [
            [sympy.cos(theta / 2), 0, 0, 1j * sympy.sin(theta / 2)],
            [0, sympy.cos(theta / 2), -1j * sympy.sin(theta / 2), 0],
            [0, -1j * sympy.sin(theta / 2), sympy.cos(theta / 2), 0],
            [1j * sympy.sin(theta / 2), 0, 0, sympy.cos(theta / 2)],
        ]
    )
    ZZ = sympy.Matrix(
        [
            [sympy.cos(theta / 2) - 1j * sympy.sin(theta / 2), 0, 0, 0],
            [0, sympy.cos(theta / 2) + 1j * sympy.sin(theta / 2), 0, 0],
            [0, 0, sympy.cos(theta / 2) + 1j * sympy.sin(theta / 2), 0],
            [0, 0, 0, sympy.cos(theta / 2) - 1j * sympy.sin(theta / 2)],
        ]
    )

    XY = sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, sympy.cos(theta / 2), 1j * sympy.sin(theta / 2), 0],
            [0, 1j * sympy.sin(theta / 2), sympy.cos(theta / 2), 0],
            [0, 0, 0, 1],
        ]
    )
    XY.simplify()
    angles = [-sympy.pi / 2, 0, sympy.pi / 5, sympy.pi / 2, sympy.pi]
    print("**" * 10)
    print("WAVEFUNCTION")
    print("-" * 10)
    print("1 qubit gates")
    print("-" * 10)
    generate_cases_1_qubit_wavefunction(Rx, "RX", angles)
    generate_cases_1_qubit_wavefunction(Ry, "RY", angles)
    generate_cases_1_qubit_wavefunction(Rz, "RZ", angles)
    generate_cases_1_qubit_wavefunction(PHASE, "PHASE", angles)
    generate_cases_1_qubit_wavefunction(RH, "RH", angles)
    print("-" * 10)
    print("2 qubit gates")
    print("-" * 10)
    generate_cases_2_qubits_wavefunction(CPHASE, "CPHASE", angles)
    generate_cases_2_qubits_wavefunction(XX, "XX", angles)
    generate_cases_2_qubits_wavefunction(YY, "YY", angles)
    generate_cases_2_qubits_wavefunction(ZZ, "ZZ", angles)
    generate_cases_2_qubits_wavefunction(XY, "XY", angles)

    print("**" * 10)
    print("EXP VALS WITHOUT ANGLES")
    print("-" * 10)
    print("1 qubit gates")
    print("-" * 10)
    generate_cases_1_qubit_exp_vals(X, "X")
    generate_cases_1_qubit_exp_vals(Y, "Y")
    generate_cases_1_qubit_exp_vals(Z, "Z")
    generate_cases_1_qubit_exp_vals(H, "H")
    generate_cases_1_qubit_exp_vals(S, "S")
    generate_cases_1_qubit_exp_vals(T, "T")
    print("-" * 10)
    print("2 qubit gates")
    print("-" * 10)
    generate_cases_2_qubits_exp_vals(CNOT, "CNOT")
    generate_cases_2_qubits_exp_vals(SWAP, "SWAP")
    generate_cases_2_qubits_exp_vals(ISWAP, "ISWAP")
    generate_cases_2_qubits_exp_vals(CZ, "CZ")

    print("**" * 10)
    print("EXP VALS WITH ANGLES")
    print("-" * 10)
    print("1 qubit gates")
    print("-" * 10)
    generate_cases_1_qubit_exp_vals_with_angles(Rx, "RX", angles)
    generate_cases_1_qubit_exp_vals_with_angles(Ry, "RY", angles)
    generate_cases_1_qubit_exp_vals_with_angles(Rz, "RZ", angles)
    generate_cases_1_qubit_exp_vals_with_angles(PHASE, "PHASE", angles)
    generate_cases_1_qubit_exp_vals_with_angles(RH, "RH", angles)
    print("-" * 10)
    print("2 qubit gates")
    print("-" * 10)
    generate_cases_2_qubits_exp_vals_with_angles(CPHASE, "CPHASE", angles)
    generate_cases_2_qubits_exp_vals_with_angles(XX, "XX", angles)
    generate_cases_2_qubits_exp_vals_with_angles(YY, "YY", angles)
    generate_cases_2_qubits_exp_vals_with_angles(ZZ, "ZZ", angles)
    generate_cases_2_qubits_exp_vals_with_angles(XY, "XY", angles)


if __name__ == "__main__":
    main()



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import os

import numpy as np

from orquestra.quantum.measurements import (
    check_parity,
    check_parity_of_vector,
    get_parities_from_measurements,
    load_parities,
    save_parities,
)
from orquestra.quantum.operators import PauliSum


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def test_parities_io():
    measurements = [(1, 0), (1, 0), (0, 1), (0, 0)]
    op = PauliSum("Z0 + Z1 + Z0*Z1")
    parities = get_parities_from_measurements(measurements, op)
    save_parities(parities, "parities.json")
    loaded_parities = load_parities("parities.json")
    assert np.allclose(parities.values, loaded_parities.values)
    assert len(parities.correlations) == len(loaded_parities.correlations)
    for i in range(len(parities.correlations)):
        assert np.allclose(parities.correlations[i], loaded_parities.correlations[i])
    remove_file_if_exists("parities.json")


def test_check_parity_odd_string():
    bitstring = "01001"
    marked_qubits = (1, 2, 3)
    assert not check_parity(bitstring, marked_qubits)


def test_check_parity_even_string():
    bitstring = "01101"
    marked_qubits = (1, 2, 3)
    assert check_parity(bitstring, marked_qubits)


def test_check_parity_odd_tuple():
    bitstring = (0, 1, 0, 0, 1)
    marked_qubits = (1, 2, 3)
    assert not check_parity(bitstring, marked_qubits)


def test_check_parity_even_tuple():
    bitstring = (0, 1, 1, 0, 1)
    marked_qubits = (1, 2, 3)
    assert check_parity(bitstring, marked_qubits)


def test_check_parity_of_vector():
    bitstrings = np.array([[0, 1, 1, 0, 1], [0, 1, 0, 0, 1]])
    marked_qubits = (1, 2, 3)
    assert np.allclose(
        check_parity_of_vector(bitstrings, marked_qubits), np.array([1, 0])
    )


def test_check_parity_of_vector_with_no_marked_qubits():
    bitstring = np.array([[1, 0], [0, 0]])
    marked_qubits = []
    assert all(check_parity_of_vector(bitstring, marked_qubits))



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Dict, List, Tuple, Union

import numpy as np

from ..distributions import MeasurementOutcomeDistribution
from ..utils import (
    bin2dec,
    dec2bin,
    sample_from_probability_distribution,
    tuple_to_bitstring,
)


def convert_integer_to_ising_bitstring(number: int, length: int) -> List[int]:
    """Converts an integer into a +/-1s bitstring (also called Ising bitstring).

    Args:
        number: positive number to be converted into its corresponding
            Ising bitstring representation
        length: length of the Ising bitstring (i.e. positive number of spins
            in the Ising system)
    Returns:
        Ising bitstring representation (1D array of +/-1)."""

    if length < 0:
        raise ValueError("Length cannot be negative.")
    binary_bitstring = dec2bin(number, length)
    ising_bitstring = [bit * 2 - 1 for bit in binary_bitstring]
    return ising_bitstring


def convert_ising_bitstring_to_integer(ising_bitstring: List[int]) -> int:
    """Converts a +/-1s bitstring (also called Ising bitstring) into an integer.
    Args:
        ising_bitstring: 1D array of +/-1.
    Returns:
        Integer number representation of the Ising bitstring."""

    binary_bitstring = [int((bit + 1) / 2) for bit in ising_bitstring]
    number = bin2dec(binary_bitstring)
    return number


def _get_random_ising_hamiltonian_parameters(
    n_spins: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """Generates random h, J, and where h and J are arrays of random coefficients
    sampled from a normal distribution with zero mean and sqrt(n_spins) sd.
    For reproducibilty, fix random generator seed in the higher level from which
    this function is called.
    Useful in the following Ising Hamiltonian: 1D Nearest Neighbor Ising Model with
    open boundary conditions, where h's are the external field values and J's
    the two-body couplings coefficients.
    Args:
        n_spins: positive number of spins in the Ising system.
    Returns:
       external_fields: n_spin coefficients sampled from a normal distribution with
        zero mean and sqrt(n_spins) sd
       two_body_couplings: n_spin x n_spin symmetric array of coefficients sampled
        from a normal distribution with zero mean and sqrt(n_spins) sd
    """
    external_fields = np.zeros((n_spins))
    two_body_couplings = np.zeros((n_spins, n_spins))
    for i in range(n_spins):
        external_fields[i] = np.random.normal(0, np.sqrt(n_spins))
        for j in range(i, n_spins):
            if i == j:
                continue
            two_body_couplings[i, j] = two_body_couplings[j, i] = np.random.normal(
                0, np.sqrt(n_spins)
            )
    return external_fields, two_body_couplings


def get_thermal_target_measurement_outcome_distribution(
    n_spins: int,
    temperature: float,
    hamiltonian_parameters: Tuple[np.ndarray, np.ndarray],
) -> MeasurementOutcomeDistribution:
    """Generates thermal states target distribution, saved in a dict where keys are
    bitstrings and values are corresponding probabilities according to the Boltzmann
    distribution formula.

    Args:
        n_spins: positive number of spins in the Ising system
        temperature: temperature factor in the boltzman distribution
        hamiltonian_parameters: values of hamiltonian parameters, namely external
            fields and two body couplings.

    Returns:
       Thermal target distribution.
       Number of positive spins in the spin state.
    """
    partition_function = 0
    external_fields, two_body_couplings = hamiltonian_parameters
    beta = 1.0 / temperature
    distribution = {}

    for spin in range(int(2**n_spins)):
        ising_bitstring = convert_integer_to_ising_bitstring(spin, n_spins)
        energy = 0
        for i in range(n_spins):
            energy -= ising_bitstring[i] * external_fields[i]
            if i != n_spins - 1:
                energy -= (
                    ising_bitstring[i]
                    * ising_bitstring[i + 1]
                    * two_body_couplings[i, i + 1]
                )
        boltzmann_factor = np.exp(energy * beta)
        partition_function += boltzmann_factor

        binary_bitstring = tuple_to_bitstring(tuple(dec2bin(spin, n_spins)))
        distribution[binary_bitstring] = boltzmann_factor

    normalized_distribution: Dict[Union[str, Tuple[int, ...]], float] = {
        key: value / partition_function for key, value in distribution.items()
    }

    return MeasurementOutcomeDistribution(normalized_distribution)


def get_thermal_sampled_distribution(
    n_samples: int,
    n_spins: int,
    temperature: float,
    hamiltonian_parameters: Tuple[np.ndarray, np.ndarray],
) -> MeasurementOutcomeDistribution:
    """Generates thermal states sample distribution
    Args:
        n_samples: the number of samples from the original distribution
        n_spins: number of spins in the Ising system
        temperature: temperature factor in the Boltzmann distribution
    Returns:
       histogram_samples: keys are binary string representations and values
        are corresponding probabilities.
    """
    distribution = get_thermal_target_measurement_outcome_distribution(
        n_spins, temperature, hamiltonian_parameters
    ).distribution_dict
    temp_sample_distribution_dict = sample_from_probability_distribution(
        distribution, n_samples
    )
    histogram_samples = np.zeros(2**n_spins)
    for samples, counts in temp_sample_distribution_dict.items():
        integer_list: List[int] = []
        for elem in samples:
            integer_list.append(int(elem))
        idx = convert_ising_bitstring_to_integer(integer_list)
        histogram_samples[idx] += counts / n_samples

    sample_distribution_dict: Dict[Union[str, Tuple[int, ...]], float] = {}
    for spin in range(int(2**n_spins)):
        binary_bitstring = tuple_to_bitstring(tuple(dec2bin(spin, n_spins)))
        sample_distribution_dict[binary_bitstring] = histogram_samples[spin]

    return MeasurementOutcomeDistribution(sample_distribution_dict)


def get_cardinality_distribution(
    n_samples: int, n_spins: int, sampled_distribution: MeasurementOutcomeDistribution
) -> List[int]:
    """Generates a list with all the occurrences associated to different cardinalities
        in a sampled distribution.

    Args:
        n_samples: the number of samples used to build the sampled distribution
            (used for normalization purposes)
        n_spins: positive number of spins in the Ising system
        sampled_distribution: measurement outcome distribution built of samples drawn
            for a target distribution

    Returns:
        cardinality_list: a list with the cardinalities of all the sampled bitstrings.

    """
    histogram_samples = np.zeros(2**n_spins)
    cardinality_list: List[int] = []
    for samples, counts in sampled_distribution.distribution_dict.items():
        integer_list: List[int] = []
        for elem in samples:
            integer_list.append(int(elem))
        idx = convert_ising_bitstring_to_integer(integer_list)
        histogram_samples[idx] += counts / n_samples
        cardinality = 0
        for elem in integer_list:
            if elem == 1:
                cardinality += 1
        for _ in range(int(counts)):
            cardinality_list.append(cardinality)

    return cardinality_list
